#open "item";;
#open "pcre";;

let print     x = print_string x; flush stdout;;
let print_err x = prerr_string x; flush stderr;;

let rec list_to_sexp x = match x with
    [] -> IUnit
  | h::tl -> (IPair (h, list_to_sexp tl));;

#infix "o";;
let prefix o f g = (fun x -> f(g(x)));;

let ip_re = regex "(\\d+)\\.(\\d+)\\.(\\d+)\\.(\\d+)";;
let parse_ip x = list_to_sexp (map (INum o float_of_string) (tl (capture ip_re x)));;

let apache_re = regex "(\\S+) - - \\[(.*?)\\] \"(.*?)\" (\\S+) (\\S+) (\\S+) (\\S+)";;
let parse_apache_log xs =
  map (fun x -> (IPair (parse_ip (hd x), list_to_sexp  (map IString (tl x)))))
    (map (fun x -> tl (capture apache_re x))
       xs);;

let census_re = regex "(\\S+), (\\S+), (\\S+), (\\S+), (\\S+), (\\S+), (\\S+), (\\S+)";;
let census_line x = match x with
    (age :: sex :: race :: union :: income :: postcode :: education :: industry :: _) ->
      list_to_sexp [INum (float_of_string age);
		    IString sex; IString race;
                    INum (float_of_string income); INum (float_of_string postcode);
                    INum (float_of_string education); INum (float_of_string industry)]
  | _ -> raise (Failure "census data parse error: incomplete line");;

let parse_census xs =
  map census_line
    (map (fun x -> tl (capture census_re x)) xs);;

let points_re = regex "(\\S+)\s+(\\S+)";;
let parse_points xs =  map (fun x -> (IPair (INum (float_of_string (hd x)), INum (float_of_string (hd(tl x))))))
  (map (fun x -> tl (capture points_re x)) xs);;

(* IDC queries *)
let queries_re = regex "(\\S+)\s+(\\S+)\s+(\\S+)\s+(\\S+)\s+(\\S+)\s+(\\S+)\s+(\\S+)\s+(\\S+)";;
let queries_line x = match x with
    (n1 :: n2 :: n3 :: n4 :: n5 :: n6 :: n7 :: n8 :: _) ->
      list_to_sexp [INum (float_of_string n1);
		    INum (float_of_string n2);
		    INum (float_of_string n3);
		    INum (float_of_string n4);
		    INum (float_of_string n5);
		    INum (float_of_string n6);
		    INum (float_of_string n7);
		    INum (float_of_string n8)]
  | _ -> raise (Failure "queries parse data error");;

let rec print_list l = match l with
  | []      -> ()
  | m :: ll -> print "--"; print m ; print "--\n"; print_list ll;;

let parse_queries xs =
  map queries_line
    (map (fun x -> tl (capture queries_re x)) xs);;

(* let parse_queries xs = *)
(*   match xs with *)
(*   | []      -> print "What!"; [INum 3.0] *)
(*   | x :: xs -> let m = capture queries_re x *)
(* 	       in match m with *)
(* 	       | [] -> print "What!!!!!"; [INum 3.0]   *)
(* 	       | _  -> print_list m;      [INum 3.0];; *)

(* let parse_student xs = *)
(*   match xs with *)
(*   | []      -> print "What!"; [INum 3.0] *)
(*   | x :: xs -> let m = capture student_re x *)
(* 	       in match m with *)
(* 	       | [] -> print "What!!!!!"; [INum 3.0] *)
(* 	       | _  -> print_list m;      [INum 3.0];; *)

(* print "Start\n"; *)
(* map student_line *)
(*   (map (fun x -> *)
(*     print "Start\n"; *)
(*     print "Line is "; *)
(*     print x; *)
(*     tl (capture student_re x)) xs);; *)

(* Student database *)
let parse_bool s = match s with
  | "t" -> true
  | "f" -> false
  | _   -> raise (Failure "student data parse error: incorrect bool");;

(* Be careful with the group limit *)
let student_re     = regex "\\s*([^;]*) ;\\s*([^;]*) ;\\s*([^;]*) ;\\s*([^;]*) ;\\s*([^;]*) ;\\s*([^;]*) ;\\s*([^;]*) ;\\s*([^;]*)";;
let student_line x = match x with
    (sid :: sname :: c1p :: c1g :: c2p :: c2g :: c3p :: c3g :: _) ->
      list_to_sexp [INum (float_of_string sid);
		    IString sname;
		    IBool (parse_bool c1p); INum (float_of_string c1g);
		    IBool (parse_bool c2p); INum (float_of_string c2g);
		    IBool (parse_bool c3p); INum (float_of_string c3g)]
  | _ -> raise (Failure "student data parse error: incomplete line");;

let parse_student xs =
  map student_line
    (map (fun x -> tl (capture student_re x)) xs);;

let text_of_file x =
  let s = open_in x in
  let rec loop () = (try let t = input_line s in t :: loop () with End_of_file -> []) in
    loop ();;

(* process commandline *)

exception Usage;;
exception Option;;

let valof x =       match x with Some y -> y | None -> raise Option;;
let usage_valof x = match x with Some y -> y | None -> raise Usage;;

(* tiny flag parsing utilities *)
type flag == string * string option ref;; (* flag name, flag argument *)

let flags = (ref [] : flag list ref);;

let flag x = let r = ref None in flags := ("--" ^ x, r) :: !flags; r;;

let process_flags () =
  let argv = tl (list_of_vect sys__command_line) in
  let rec go (x : string list) = match x with
      f :: arg :: tl ->
        (try assoc f (!flags) := Some arg with Not_found -> raise Usage);
        go tl
    (* Hack for IUnit *)
    | f :: [] ->
      (if f = "--unit" then
	try assoc f (!flags) := Some "" with Not_found -> raise Usage;
      else
	raise Usage);
      []
    | _ -> x
  in go argv;;

let rec attempt x =
  match x with
    []        -> raise Usage
  | (h :: tl) -> try h() with Usage -> attempt tl;;

let dev_random_poll () =
  (* In linux this is high quality noise, but blocks *)
  (* print "[Info] Initializing RNG Generator [need some entrophy, do some IO ;)] ..."; *)
  (* let input = io__open_in "/dev/random" in *)
  print "[Warning] Testing mode: Using low quality noise from pseudo random generator...";
  let input = io__open_in "/dev/urandom" in
  let res   = io__input_binary_int input in
  io__close_in input;
  print "... done\n";
  res;;

(* Placeholder for the protected procedure *)
let init_protected = ();;

(* returns the db to call the main function *)
let init query =
  let _ = random__init (dev_random_poll ()) in
  let apache_log = flag "apache-log" in
  let points     = flag "points"     in
  let census     = flag "census"     in
  let student    = flag "student"    in
  let queries    = flag "queries"    in
  let unit       = flag "unit"       in
  let _ = process_flags ()           in
   (* get database *)
  let db = attempt [
    (fun () -> IBag (parse_apache_log (text_of_file (usage_valof !apache_log)))) ;
    (fun () -> IBag (parse_points     (text_of_file (usage_valof !points)))) ;
    (fun () -> IBag (parse_census     (text_of_file (usage_valof !census)))) ;
    (fun () -> IBag (parse_student    (text_of_file (usage_valof !student)))) ;
    (fun () -> IBag (parse_queries    (text_of_file (usage_valof !queries)))) ;
    (fun () -> let _ = (usage_valof !unit) in IUnit) ; ] in
   (* run query *)
  let res = iapp query db in
  print (item_to_string ((fuzz_of_item res) ()));
  print "\n";
  ();;

let main query =
  printexc__f
    (fun () ->
      (try init query
       with
	 Usage -> (print_err ("Usage: ./runquery --apache-log [logfile]\n" ^
                              "                  --points  [pointsfile]\n" ^
                              "                  --census  [censusfile]\n" ^ 
                              "                  --student [studentfile]\n" ^ 
                              "                  --queries [queryfile]\n" ^ 
			      "                  --unit\n")
                  ;
                   exit 0)
       | Failure s -> (print_err ("Error: " ^ s ^ "\n")))) ()
;;
