#open "item";;
#open "math";;

(* Some util *)
let print x = print_string x; flush stdout;;
let print_err x = prerr_string x; flush stderr;;

(* Standard embedding themselves *)
let iapp x y = match x with
    IFn f ->
      begin
	match y with
	| _ -> f y
      end
  | _ -> raise (Failure "runtime type error: applying a nonfunction");;

let iequal x y = IBool (x = y) ;;
let igt x y    = IBool (x > y) ;;
let ilt x y    = IBool (x < y) ;;

let iconj x y  = IBool (
  match (x, y) with
    (IBool true, IBool true) -> true
  | _ -> false);;

let idisj x y  = IBool (
  match (x, y) with
    (IBool false, IBool false) -> false
  | _ -> true);;

let iplus x y = match (x, y) with
    (INum xx, INum yy) -> INum (xx +. yy)
  | (IString xx, IString yy) -> IString (xx ^ yy)
  | _ -> raise (Failure "runtime type error: adding nonnumbers");;
let iminus x y = match (x, y) with
    (INum xx, INum yy) -> INum (xx -. yy)
  | _ -> raise (Failure "runtime type error: subtracting nonnumbers");;
let idiv x y = match (x, y) with
    (INum xx, INum yy) -> INum (xx /. yy)
  | _ -> raise (Failure "runtime type error: dividing nonnumbers");;
let imult x y = match (x, y) with
    (INum xx, INum yy) -> INum (xx *. yy)
  | _ -> raise (Failure "runtime type error: dividing nonnumbers");;

let itrue x    = x = IBool true;;

(* Monad *)
let return =
  IFn (fun x -> IFuzz (fun () -> x));;

let dethunk x = match x with
    IFuzz f -> f ()
  | _ -> raise (Failure "runtime type error: dethunking a nonthunk");;

let sample k f = k (dethunk f);;

let fuzz_of_item x = match x with
    IFuzz y -> y
  | _ -> raise (Failure "runtime error, expected fuzz");;

(* Util *)
let num2string = IFn (fun x -> IString (
  match x with
    INum xx -> string_of_float xx
  | _ -> raise (Failure "runtime type error: num2string on nonnumber")));;

let rec item_to_string x = match x with
    INum n      -> string_of_float n
  | IPair(a, b) -> "(" ^ (item_to_string a) ^ "," ^ (item_to_string b) ^ ")"
  | IUnit       -> "()"
  | IInl a      -> "(Left " ^ (item_to_string a) ^ ")"
  | IInr a      -> "(Right " ^ (item_to_string a) ^ ")"
  | IString s   -> s
  | IBool b     -> (if b then "true" else "false")
  | IBag _      -> "Bag"
  | IMemoryBag n -> "MemBag of " ^ (string_of_int n)
  | IFinMap   _ -> "Finite Map"
  | IFn _       -> "Function!"
  | IFuzz _     -> "Fuzz value";;


(* Clip and fuzz *)
let clip = IFn (fun x -> match x with
    INum x -> INum (if x >. 1.0 then 1.0 else if x <. 0.0 then 0.0 else x)
  | _      -> raise (Failure "runtime type error: clipping a nonnumber"));;

(* fuzz with fixed epsilon *)
(* fuzz: epsilon -o num -o fuzzy num  *)
let fuzz = fun eps -> IFn (fun x -> IFuzz (fun () -> iplus x (INum (lap (1.0 /. eps)))));;

(* lapfuzz : Num -o Num -o fuzzy Num *)
(* let lapfuzz = IFn ( fun k -> IFn (fun x -> IFuzz (fun () -> iplus x (INum (lap (k))))));; *)

(* this version of fuzz is a laplace noise function. Typically k will be 1/epsilon*)

let icswp x = match x with
    IPair (x, y) -> IPair (if x <= y then (x, y) else (y, x))
  | _            -> raise (Failure ("runtime type error, cswp: " ^ item_to_string x));;

let decons_pair x = match x with
    IPair p -> p
  | _       -> raise (Failure ("runtime type error, deconstructing as pair: " ^ item_to_string x));;




