#!/usr/bin/python
# -*- coding: utf-8 -*-

from z3 import *
from z3rinf import *

# Util
def z3_rat_to_float(s):
    if isinstance(s, RatNumRef):
        n = s.numerator().as_long()
        d = s.denominator().as_long()

        # Alternative to "from __future__ import division"
        return n / float(d)
    else:
        return s

def z3tf(s):
    return z3_rat_to_float(s)

def svar_names(lvars):
    s = ""
    for sv in lvars:
        s = s + " " + sv.name
    return s

def print_svar(model, svar):
    if is_true(model[svar.inf]):
        print "%s = ∞" % svar.name
    else:
        svar_val = z3_rat_to_float(model[svar.val])
        print "%s = %.8f" % (svar.name, svar_val)

def print_model(model, svars):
    for svar in svars:
        print_svar(model, svar)

# Check a list of constraints.
def check(F):
    # Create the solver instance.
    solver = Tactic('qfnra-nlsat').solver()
    solver.add(F)

    r = solver.check()

    # A model only exists if the constraints are satisfiable
    if r == unsat or r == unknown:
        return r
    else:
        return solver.model()

c_iter  = 0

# Parameters controlling minimization error
use_abs = True
min_abs_error = 0.000001
min_rel_error = 0.001

def print_error():
    s = ""

    if use_abs:
        s = " [up to %.2f absolute error]" % min_abs_error
    else:
        s = " [up to %.2f%% error]" % (min_rel_error * 100)
    return s

# This controls when to stop iterating
def error_ok(val, lb, up):
    global min_abs_error, min_rel_error, use_abs

    abs_error = z3tf(simplify((up-lb)*2));
    rel_error = z3tf(simplify((up-lb)/val));

    # print abs_error, rel_error

    # Stop when the error is less than 0.01 (absolute)
    if use_abs:
        return (abs_error > min_abs_error)
    else:
        return (rel_error > min_rel_error)

def min_var(solver, svar):

    last_model = solver.model()

    # We first forbid the variable to be infinity
    # This is unconditional
    solver.push()
    solver.add(Not(svar.inf))

    # If X got assigned infinity in the model, try to get an upper
    # bound.
    if is_true(last_model[svar.inf]):
        # Check for s < infty
        r = solver.check()

        if r == unsat or r == unknown:
            # Can't be less than infinity, we stop and return the old solver
            if verbose:
                print "Variable is infinity and cannot be minimized"
            solver.pop()
            return solver
        else:
            # We found a model with v less than infinity
            last_model = solver.model()
            if verbose:
                print "Variable was infinity but found new value %s" % str(last_model[X.val])

    lowBound = RealVal(0.0)
    upBound  = last_model[svar.val]

    global c_iter
    firstfail = False

    # Minimization loop.

    # XXX: Fix this
    # Loop invariant, the current solver object contains the model corresponding to the bounds.
    while error_ok(last_model[svar.val], lowBound, upBound):
        c_iter = c_iter + 1

        # Debug
        if verbose:
            print "Iteration %d for %s current value %f, l-bound: %f, u-bound: %f" % (c_iter, svar.name, z3tf(last_model[svar.val]), z3tf(lowBound), z3tf(upBound))

        # Try a new upperbound
        curr = simplify((upBound+lowBound)/2)

        # Couldn't make sense of this yet
        # curr = simplify((9*lowBound+upBound)/10)

        if verbose:
            print "New upperBound candidate for %s: %f" % (svar.name, z3tf(curr))

        solver.push()
        solver.add(svar.val <= curr)
        r = solver.check()

        if r == unsat:
            # The upper bound was too low.
            if verbose:
                print "The upper bound was too low, trying a greater one..."

            solver.pop()

            # Check if the upper bound can be improved at all!
            solver.push()
            solver.add(svar.val < upBound)
            r = solver.check()
            solver.pop()

            if r == unsat:
                if verbose:
                    print "Upper bound cannot be improved, finish with value %.2f" % z3tf(upBound)
                break
            else:
                # Yes it can be improved, make the new lower bound the failed upper one.
                lowBound = curr
        else:
            if verbose:
                print "New upper bound found, trying a lower one..."

            last_model = solver.model()
            upBound = curr
            solver.pop()

    # We return the solver with the added minimization constraints.
    solver.add(svar.val <= upBound)
    return solver

# Minimize variables
def min(cs, varlist):
    # A very strong hammer, but has no effect compared to
    # constraint.simpl_subst_cs:

    # solver = AndThen(Tactic('fm'),
    #                  Tactic('propagate-values'),
    #                  Tactic('propagate-ineqs'),
    #                  Tactic('der'),
    #                  Tactic('ctx-simplify'),
    #                  Tactic('ctx-solver-simplify'),
    #                  Tactic('simplify'),
    #                  Tactic('solve-eqs'),
    #                  Tactic('elim-and'),
    #                  Tactic('sat-preprocess'),
    #                  Tactic('aig'),
    #                  Tactic('vsubst'),
    #                  Tactic('fm'),
    #                  Tactic('propagate-values'),
    #                  Tactic('propagate-ineqs'),
    #                  Tactic('der'),
    #                  Tactic('ctx-simplify'),
    #                  Tactic('ctx-solver-simplify'),
    #                  Tactic('simplify'),
    #                  Tactic('solve-eqs'),
    #                  Tactic('elim-and'),
    #                  Tactic('sat-preprocess'),
    #                  Tactic('qfnra-nlsat')).solver()

    solver = Tactic('qfnra-nlsat').solver()
    # solver.set(randomize=False)

    solver.add(cs)

    r = solver.check()

    if r == unsat or r == unknown:
       return r

    for svar in varlist:
        if verbose:
            print "Starting minimization on variable %s" % svar.name
            print "Initial model for the variables to minimize:"
            print_model(solver.model(), varlist)

        solver = min_var(solver, svar)
        # This is a minor pitfall for now
        solver.check()

    return solver.model()

import inspect

# We remove the last constraint from the list, and stop once the
# model is satifiable. The index will give us the solution.
def blame_constraint(n_cs, cs_list):
    try:
        cs_list.pop()
        n_cs  = n_cs + 1
        model = check(cs_list)

        if model == unsat:
            return blame_constraint(n_cs, cs_list)
        elif model == unknown:
            print "Model Unknown, Ouch!"
            return Error
        else:
            return n_cs
    except IndexError:
        print "No model but no constraint to blame! Something is seriously broken!"

import time

def start_timing():
    return time.clock()

def stop_timing(z3_time):
    elapsed = (time.clock() - z3_time)

    if len(fuzz.minimize) > 0 :
        min_str = "[Completed %d iters, CPU per iteration %.3f (avg)]" % (c_iter, elapsed/(c_iter+1))
    else:
        min_str = ""

    print "Total CPU time:", elapsed, "seconds", min_str

# Command line parsing
import sys

verbose = False

if len(sys.argv) < 2:
    print "Fuzz-generated constraints file missing"
else:
    fdir, fname = os.path.split(sys.argv[1])
    mname, ext  = os.path.splitext(fname)

    sys.path.append(fdir)

    if len(sys.argv) == 3 and sys.argv[2] == "-v":
        verbose = True

    try:
        fuzz = __import__(mname)

        # Pass variable to minimize in this case r2
        z3_time = start_timing()

        if len(fuzz.minimize) == 0:
            model = check(fuzz.constraints)
        else:
            print "Minimizing on" + svar_names(fuzz.minimize) + print_error()
            model = min(fuzz.constraints, fuzz.minimize)

        stop_timing(z3_time)

        if model == unsat:
            print "Cannot find a model, typechecking fails"
            # TODO: Implement iterative searching of the blaming constrant.
            cs_removed = blame_constraint(0, fuzz.constraints)
            cs_str_index = len(fuzz.cs_string) - cs_removed

            print "The non-satisfiable constraint was generated at " + str(fuzz.cs_string[cs_str_index])
        elif model == unknown:
            print "Model Unknown, Ouch!"
        else:
            # Z3 built-in printing, use only for debug
            # print model

            # This prints all variables.
            if verbose:
                print "Verbose information: all the variables"
                print_model(model, fuzz.svars)

            # This is the standard operating mode:
            if len(fuzz.interesting_svars) > 0:
                print "Inferred sensitivities for user variables:"
                print_model(model, fuzz.interesting_svars)

            print "TypeChecking Completed: Pass"

    except ImportError:
        print "Error importing module %s %s %s %s" % (fdir, fname, mname, ext)
