open Format

open Ctx
open Syntax
open Print

(*
  We generate output according to the syntax defined in item.mli:

type item =
            | INum of float
            | IInl of item
            | IInr of item
            | IFn of item -> item
            | IPair of item * item
            | IFuzz of unit -> item

            | IUnit
            | IBag of int
            | IBool of bool
            | IString of string
	    | NULL;;

Plus some primitives.

Only in npitem.ml:
	    | IFinMap of ((item * item) list)

*)

let paren s = "(" ^ s ^ ")"

let header =
"#open \"item\";;
#open \"bag\";;
#open \"init\";;

let p_inl = IFn (fun x -> IInl x);;
let p_inr = IFn (fun x -> IInr x);;
let tensor_pair = IFn (fun x -> IFn (fun y -> IPair (x, y)));;

let wrap1 f = IFn (fun x -> f x);;
let wrap2 f = IFn (fun x -> IFn (fun y -> f x y));;

let op_add = wrap2 iplus;;
let op_sub = wrap2 iminus;;
let op_eq = wrap2 iequal;;
let op_lt = wrap2 ilt;;
let op_gt = wrap2 igt;;
let op_mul = wrap2 imult;;
let op_div = wrap2 idiv;;

let op_iadd = wrap2 iplus;;
let op_isub = wrap2 iminus;;

let op_land = wrap2 iconj;;
let op_lor  = wrap2 idisj;;

let cswp = wrap1 icswp;;

let div2  = IFn (fun x -> iapp (iapp (wrap2 idiv) x) (INum 2.0));;
let div3  = IFn (fun x -> iapp (iapp (wrap2 idiv) x) (INum 3.0));;

(* In the old style iplus takes care of both *)
let op_string_concat = IFn (fun s1 -> IFn (fun s2 ->
    match s1 with
    | IString ss1 ->
      (match s2 with
      | IString ss2 -> IString (ss1 ^ ss2)
      | _ -> raise (Failure \"what happened to our typechecker\"))
    | _ -> raise (Failure \"what happened to our typechecker\")));;

let p_sample = IFn (fun x -> IFn (fun f ->
  match f with
  | IFn f' -> sample f' x
  | _ -> raise (Failure \"what happened to our typechecker\") ));;

(* fix f x = f (fix f) x *)
let rec p_fix =
  IFn (fun f -> IFn (fun x ->
    iapp (iapp f (iapp p_fix f)) x
  ));;

let if_then_else = IFn (fun test -> IFn (fun opts ->
  match opts with
  | IPair (o1, o2) ->
    (match test with
    | IBool(b) -> if b then (iapp o1 IUnit) else (iapp o2 IUnit)
    | _ -> raise (Failure \"what happened to our typechecker\"))
  | _ -> raise (Failure \"what happened to our typechecker\") 
));;

let query =
"

let body =
" in
main query;;"

let rec gen_primitive prim =
  match prim with
    PrimTUnit        -> "IUnit"
  | PrimTNum(r)      -> paren ("INum "  ^ string_of_float r)
  | PrimTInt(n)      -> paren ("INum "  ^ string_of_int   n ^ ".0" )
  | PrimTBool(b)     -> paren ("IBool " ^ string_of_bool  b)
  | PrimTString(s)   -> paren ("IString " ^ "\"" ^ s ^ "\"")
  | PrimTFun(f, _ty) -> f

let rec gen_term ppf t =
  match t with
      (* TODO: not very happy using debug information for this, generate
	 names from the index in the future *)
      TmVar (_, v)  -> fprintf ppf "%s" v.v_name

    | TmFuzz (_, n) -> fprintf ppf "(fuzz %f)" n

    (* Will be represented as applications soon *)
    | TmTensDest (_,  b_x, b_y, tm_e1, tm_e2) ->
      fprintf ppf "(match (decons_pair %a) with (%a,%a) ->@\n @[%a@])"
	gen_term tm_e1 pp_binfo b_x pp_binfo b_y gen_term tm_e2

    | TmUnionCase (_, tm_e, bi_l, tm_l, bi_r, tm_r) ->
      fprintf ppf "(match %a with @[<v>| IInl %a -> %a @,| IInr %a -> %a@,| _ -> %s)@]"
	gen_term tm_e pp_binfo bi_l gen_term tm_l pp_binfo bi_r gen_term tm_r
	"raise (Failure \"what happened to our typechecker\")"

    (* Regular stuff for fuzz *)
    | TmPrim (_, prim) -> fprintf ppf "%s" (gen_primitive prim)

    | TmApp (_, f, a)  -> fprintf ppf "(iapp @[%a@\n%a@])" gen_term f gen_term a

    | TmAbs (_, b, _sty, _t2, body) ->
      fprintf ppf "(IFn (fun %s ->@\n @[%a@]))" b.b_name gen_term body

    | TmAmpersand (_i, e1, e2) -> fprintf ppf "(IPair (%a,%a))" gen_term e1 gen_term e2

    | TmUnfold (_i, tm) -> gen_term ppf tm
    | TmFold(_i, _, tm) -> gen_term ppf tm

    (* let bi = e1 in e2 *)
    | TmLet (_, bi, _ty, e1, e2) ->
      (* It is pointless to generate code for primitives *)
      begin
	match e1 with
	| TmPrim(_, PrimTFun(_, _)) -> gen_term ppf e2
	| _            ->
	  fprintf ppf "(let %a = %a in@\n %a)" pp_binfo bi gen_term e1 gen_term e2
      end
    | TmTypedef (_i,_tn,_ty,tm) -> gen_term ppf tm

    (* TODO: DFuzz *)
    | TmListCase (_) -> fprintf ppf "XXX TODO list case"
    | TmNatCase (_)  -> fprintf ppf "XXX TODO nat case"

    | TmUnpack(_,_,_,_,_)    -> fprintf ppf "XXX TODO: TmUnpack"
    | TmPack(_,_,_,_)        -> fprintf ppf "XXX TODO: TmPack"
    | TmSiApp(_,_,_)         -> fprintf ppf "XXX TODO: TmInst"
    | TmSiAbs(_,_,_,_)         -> fprintf ppf "XXX TODO: TmInst"

let rec gen_program ppf t =
  fprintf ppf "%s@\n" header;
  gen_term ppf t;
  fprintf ppf "%s@." body

