(*
  Fuzz compiler.

  Copyright: The University of Pennsylvania 2012
  License: ...

  Constraint engine.

*)

open Format

open Print
open Syntax
open Support

(* We do some constraint housekeeping here before submitting them to
   the actual constraint solver *)

type si_cstr =
  CSiEq of FileInfo.info * si * si
| CSiGe of FileInfo.info * si * si

(* Simple accumulating constraint solver *)
type c_store = si_cstr list

let empty_cs = []

(* Peephole optimization of sensitivities and constraints *)

(* Simplify sensitivities according to these rules that are sound for
   our constraint solver:

   a + inf = inf
   0 + a   = a
   0 * a   = 0
   a * inf = inf (if a != 0) and a not a variable
   1 * a   = a
*)
let rec si_simpl si = match si with
  | SiAdd(si1, si2)  ->
    let si1' = si_simpl si1 in
    let si2' = si_simpl si2 in
    if (si1' = SiInfty) || (si2' = SiInfty) then
      SiInfty
    else if si1' = SiConst(0.0) then
      si2'
    else if si2' = SiConst(0.0) then
      si1'
    else
      SiAdd(si1', si2')

  | SiMult(si1, si2) ->
    let si1' = si_simpl si1 in
    let si2' = si_simpl si2 in
    if (si1' = SiInfty || si2' = SiInfty) then
      SiInfty
    else if si1' = SiConst(1.0) then
      si2'
    else if si2' = SiConst(1.0) then
      si1'
    else
      SiMult(si1', si2')
  | _ -> si

let cs_eq i si1 si2 = CSiEq(i, si_simpl si1, si_simpl si2)
let cs_ge i si1 si2 = CSiGe(i, si_simpl si1, si_simpl si2)

(* None   -> we can't decide the constraint,
   Some b -> we know the constraint is b
*)
let cs_decide si_cs = match si_cs with
  (* Simple cases *)
  | CSiGe(_, _, SiConst(0.0)) -> Some true
  | CSiGe(_, SiInfty, _)      -> Some true

  (* Ground values can be decided here *)
  | CSiEq(_, SiConst(x), SiConst(y))        -> Some (x = y)
  | CSiGe(_, SiConst(x), SiConst(y))        -> Some (x >= y)

  | CSiEq(_, SiConst(_), SiInfty)           -> Some false
  | CSiEq(_, SiInfty, SiConst(_))           -> Some false

  | CSiGe(_, SiConst(_), SiInfty)           -> Some false

  (* More general eq test *)
  | CSiEq(_, si1, si2)        -> if si1 = si2 then Some true else None

  (* Forward to the constraint solver *)
  | _                         -> None

(* We don't do stuff like 2 * (2 * x) -> 4 * x so we can track some of
   the applications better *)

let post_cs cstore si_cs =
  match cs_decide si_cs with
  | None     -> Some (cstore@[si_cs])
  | Some res -> if res then Some cstore else None

(* To be deprecated *)
let si_cs_simpl si_cs = match si_cs with
  | CSiEq(i, si1, si2) -> CSiEq(i, si_simpl si1, si_simpl si2)
  | CSiGe(i, si1, si2) -> CSiGe(i, si_simpl si1, si_simpl si2)

let cs_map f cs = match cs with
  | CSiEq(i, si1, si2) -> CSiEq(i, f si1, f si2)
  | CSiGe(i, si1, si2) -> CSiGe(i, f si1, f si2)

let cs_fold f_eq f_ge cs = match cs with
  | CSiEq(_i, si1, si2) -> f_eq si1 si2
  | CSiGe(_i, si1, si2) -> f_ge si1 si2

(* Reimplement with fold? This will be needed in the dependent case *)
let rec cs_separate cs = match cs with
  | []                           -> ([], [])
  | CSiEq(_, s1, s2)      :: csl -> let (cs_eq, cs_ge) = cs_separate csl in ((s1, s2)::cs_eq, cs_ge)
  | CSiGe(_, _, _) as cse :: csl -> let (cs_eq, cs_ge) = cs_separate csl in (cs_eq          , cse::cs_ge)

let cs_efv cs = let f = (fun s1 s2 -> VarSet.union (si_efv s1) (si_efv s2)) in
		cs_fold f f cs

(* Occurs check! *)
let cs_oc i cs =
  let fv_set = cs_efv cs in
  VarSet.mem i fv_set

(* No binders for the moment *)
let cs_subst i si cs = cs_map (si_esubst 0 i (fun _ -> si)) cs

(* Eliminate all the equalities of the form "var = sens" from the
   constraint store *)
(* XXX perform occurs check! *)
let simpl_subst_cs (si1, si2) cs =
  let sub =
    match (si1, si2) with
    | SiEVar v1, SiEVar v2 ->
      if v1.v_index < v2.v_index then
	cs_subst v2.v_index (SiEVar v1)
      else
	cs_subst v1.v_index (SiEVar v2)
    | SiEVar v, t -> cs_subst v.v_index t
    | t, SiEVar v -> cs_subst v.v_index t
    | _, _        -> (fun x -> x)
  in List.map sub cs

let simpl_subst_eq (si1, si2) eql =
  let my_sisub i t          = si_esubst 0 i (fun _ -> t)         in
  let pair_sub i t (s1, s2) = (my_sisub i t s1, my_sisub i t s2) in
  let sub =
    match (si1, si2) with
    | SiEVar v1, SiEVar v2 ->
      if v1.v_index < v2.v_index then
	pair_sub v2.v_index (SiEVar v1)
      else
	pair_sub v1.v_index (SiEVar v2)
    | SiEVar v, t -> pair_sub v.v_index t
    | t, SiEVar v -> pair_sub v.v_index t
    | _, _        -> (fun x -> x)
  in List.map sub eql

(* XXX: Check that eq is eliminable, for instance X * Y = X + Y should
   not be eliminated *)
let rec cs_subst_list eq_l cs = match eq_l with
    []      -> cs
  | eq::eql ->
    let cs_sub = simpl_subst_cs eq cs  in
    let eq_sub = simpl_subst_eq eq eql in
    cs_subst_list eq_sub cs_sub

(* XXX: This still is not safe but for restricted subset, use with care *)
let simpl_subst cs =
  let (cs_eq, cs_ge)       = cs_separate cs in
  cs_subst_list cs_eq cs_ge

(*
   More simpl:

   - partition the store in eq constraints and not eq
   - substitute (with occurs check!)
   - simplify ground diseqs
   - ...
   - Profit!

*)

(* Pretty printing *)

(***********************************************************************)
(* Pretty printing for constraint stores *)

let pp_fi = FileInfo.pp_fileinfo

let pp_si_cstr_elem fmt si_cstr = match si_cstr with
    CSiEq(_, si1, si2) -> fprintf fmt "%a =  %a" pp_si si1 pp_si si2
  | CSiGe(_, si1, si2) -> fprintf fmt "%a >= %a" pp_si si1 pp_si si2

let pp_si_cstr_elem_fi fmt si_cstr = match si_cstr with
    CSiEq(i, si1, si2) -> fprintf fmt "%a   %a =  %a" pp_fi i pp_si si1 pp_si si2
  | CSiGe(i, si1, si2) -> fprintf fmt "%a   %a >= %a" pp_fi i pp_si si1 pp_si si2

let pp_si_cstr = (pp_list pp_si_cstr_elem_fi)

let pp_si_store fmt cs =
  fprintf fmt "%a" pp_si_cstr cs

(* Move uninteresting constraints to the end *)
let uninteresting_cs cs =
  match cs with
  | CSiGe(_, SiEVar _, SiConst 1.0) -> true
  | _ -> false

let clean l       = List.partition (uninteresting_cs) l
let clean_simpl l = clean (List.map si_cs_simpl l)

let pp_cs_name fmt cs = match cs with
  | CSiGe(_i, SiEVar v, SiConst 1.0) -> pp_vinfo fmt v
  | _ -> fprintf fmt "*!"

let pp_cs_dirty fmt cs = fprintf fmt "@[<hov>[%a] >= 1.0@]" (pp_list pp_cs_name) cs

(**********************************************************************)
(* Below are useful bits for the future (not used for now) *)

(* For now we use a simple structure for the constraint solver: A list
   of equations. The checking procedure just goes linearly on all all
   the constraints, decomposing them, doing occurs checks and applying
   substitutions *)

(* Efficient constraint solver:

 - The constraint store is a hashmap of substitions (variable,
   subst), where subst is either "Free", a pointer to other variable
   or a type term representing "v -> t".

 - When a constraint is posted, it gets decomposed to primitive
   substitutions. We detect failure for a lot of cases here.  We apply
   the substitution and continue decomposing the remaining equations.

 - Occurs check guarantees that the process terminates.
*)

(**********************************************************************)
(* Failure handling *)
(* We use a monadic style to notify non-satisfaction. It forces the
   user to handle it.
*)

let fail      = None
let return x  = Some x
let (>>=) m f = match m with
    None   -> None
  | Some x -> f x

let iwhen b s = if b then Some s else None

(* Native @@ is already in ocaml 4.0 *)
let (@@) x y = x y

(**********************************************************************)
(* Constraint Store *)

type ty_cstr =
  CTyEq   of ty * ty
| CTyInst of ty * ty


(* The empty store *)
let ctrue = ([], [])

(**********************************************************************)
(* Operations on constraint stores *)

(* Lift @ to the product == merging for our fake constraint store *)
let merge_cs (l1,r1) (l2,r2) = (l1@l2, r1@r2)

(* Check weather the constraint store contains coherent constraints -- fake *)

(* We need to perform checking here *)

(* replace with fold *)
let rec merge_cs_list csl = match csl with
    []      -> ctrue
  | (x::xs) -> merge_cs x (merge_cs_list xs)

(**********************************************************************)
(* Constraint Manipulation *)

let rec decompose_sieq i asi1 asi2 = match (asi1, asi2) with
    SiVar _ as v, t            -> return ([], [CSiEq (i, v, t)])
  | t, (SiVar _ as v)          -> return ([], [CSiEq (i, v, t)])

  | SiInfty, SiInfty           -> return ctrue
  | SiIntConst i, SiIntConst j -> iwhen (i = j) ctrue
  | SiConst    i, SiConst    j -> iwhen (i = j) ctrue

  | SiAdd(sil1, sil2), SiAdd(sir1, sir2)    ->
    decompose_binary_si i sil1 sil2 sir1 sir2

  | SiMult(sil1, sil2), SiMult(sir1, sir2)  ->
    decompose_binary_si i sil1 sil2 sir1 sir2

  (* Fail otherwise *)
  | _, _ -> fail

and decompose_binary_si i l1 l2 r1 r2 =
  decompose_sieq i l1 r1                   >>= fun eqsl ->
  decompose_sieq i l2 r2                   >>= fun eqsr ->
  return @@ merge_cs eqsl eqsr

(* THINK: Should we decouple descomposition from checking? *)
let rec decompose_tyeq i aty1 aty2 = match (aty1, aty2) with
    TyVar _ as v, t      -> return ([CTyEq (v, t)] , [])
  | t, (TyVar _ as v)    -> return ([CTyEq (v, t)] , [])

  | TyPrim p1, TyPrim p2 -> iwhen (p1 = p2) ctrue

  | TyPrim1(p1, ty1), TyPrim1(p2, ty2) ->
    iwhen (p1 = p2) ()                          >>= fun _ ->
    decompose_tyeq i ty1 ty2

  | TyUnion(tyl1, tyl2), TyUnion(tyr1, tyr2) ->
    decompose_binary_ty i tyl1 tyl2 tyr1 tyr2

  | TyTensor(tyl1, tyl2), TyTensor(tyr1, tyr2) ->
    decompose_binary_ty i tyl1 tyl2 tyr1 tyr2

  | TyAmpersand(tyl1, tyl2), TyAmpersand(tyr1, tyr2) ->
    decompose_binary_ty i tyl1 tyl2 tyr1 tyr2

  | TyLollipop(tyl1, sil, tyl2), TyLollipop(tyr1, sir, tyr2) ->
    decompose_binary_ty i tyl1 tyl2 tyr1 tyr2     >>= fun eqs  ->
    decompose_sieq i sil  sir                     >>= fun eqsi ->
    return @@ merge_cs eqs eqsi

  | TyMu(_, ty1), TyMu(_, ty2)    -> decompose_tyeq i ty1 ty2
  | _, _ -> fail

and decompose_binary_ty i l1 l2 r1 r2 =
  decompose_tyeq i l1 r1                   >>= fun eqsl ->
  decompose_tyeq i l2 r2                   >>= fun eqsr ->
  return @@ merge_cs eqsl eqsr

let free_vars _ty = []

let occurs_check _v _t = false

let apply_subst _v _t rs = rs

(* Warning!!! *)
let check_cs (l1,r1) = Some (l1, r1)

(**********************************************************************)
(* Constraint Telling / Posting *)

(**********************************************************************)
(* Constraints over types *)

(* The main type equality constraint *)
let post_ty_eq i cs ty1 ty2 =
  decompose_tyeq i ty1 ty2                       >>= fun eqs ->
  return @@ merge_cs cs eqs

(* The aux is just to maintain a particular order *)
let rec post_ty_eq_list_aux i ty tyl = match tyl with
    []          -> return ctrue                (* Should return cs but anyways... *)
  | ty' :: tyll ->
    decompose_tyeq i ty ty'                      >>= fun eqs ->
    post_ty_eq_list_aux i ty tyll                >>= fun eqsl ->
    return @@ merge_cs eqs eqsl

(* Helper for lists of types *)
let post_ty_eq_list i cs ty tyl =
  post_ty_eq_list_aux i ty tyl                   >>= fun eqs ->
  return @@ merge_cs cs eqs

(* Helper for lists of types against list of types *)

let rec post_ty_eq_llist_aux i tyl1 tyl2 =
  match tyl1 with
    [] -> return ctrue
  | ty :: tyl ->
    post_ty_eq_llist_aux i tyl tyl2                  >>= fun eqsl ->
    post_ty_eq_list_aux  i ty  tyl2                  >>= fun eqs1 ->
    return @@ merge_cs eqs1 eqsl

let post_ty_eq_llist i cs tyl1 tyl2 =
  post_ty_eq_llist_aux i tyl1 tyl2               >>= fun eqs ->
  return @@ merge_cs cs eqs

(* The type equality constraint for recursive types *)
let post_ty_mu_eq i cs ty1 ty2 =
  decompose_tyeq i ty1 ty2                       >>= fun eqs ->
  return @@ merge_cs cs eqs

(* Helper function for the unfold of recursive types *)
(* It takes a type, checks that it is a mu type and returns its unfold *)
let post_is_mu cs ty1 =
  return @@ (cs, ty1)

(* Pretty printing *)
let pp_ty_cstr_elem fmt ty_cstr = match ty_cstr with
    CTyEq  (ty1, ty2) -> fprintf fmt "%a = %a"  pp_type ty1 pp_type ty2
  | CTyInst(ty1, ty2) -> fprintf fmt "%a :< %a" pp_type ty1 pp_type ty2

let pp_ty_cstr = (pp_list pp_ty_cstr_elem)

let pp_c_store fmt cs =
  let (ty_cs, si_cs) = cs in
  fprintf fmt "Type constrs: @[<v>%a@]@\nSens constrs: @[<v>%a@]" pp_ty_cstr ty_cs pp_si_cstr si_cs

