open Support.Error

open Syntax

(* ---------------------------------------------------------------------- *)
(* Context management *)

(* Contexts of type 'a *)
type 'a ctx = (var_info * 'a) list

(* Term and type variables *)
type context =
    {
      var_ctx   : (si option * ty) ctx;
      tyvar_ctx : kind ctx;
    }

let length ctx = List.length ctx

let empty_context = { var_ctx = []; tyvar_ctx = []; }

(* Return the index of a binding if it exists. Let the caller handle
   the error *)
let rec slookup id ctx =
  match ctx with
      []                -> None
    | (var, value) :: l ->
      if var.v_name = id then
	Some (var, value)
      else
	slookup id l

let lookup_var id ctx =
  slookup id ctx.var_ctx

let lookup_tyvar id ctx =
  slookup id ctx.tyvar_ctx

(* Shift a context and its contents *)

(* When we introduce a type var, all type variables in the term
   bindings must be shifted *)
let rec shift_type_var_ctx ctx n = match ctx with
    []                  -> []
  | (v, (si, ty)) :: l -> (v, (Option.map (si_shift n) si, shift_type n ty)) :: shift_type_var_ctx l n

(* When we introduce a term var, all term variables must be shifted *)
let rec shift_term_var_ctx ctx n = match ctx with
    []          -> []
  | (v, b) :: l -> (shift_var n v, b) :: shift_term_var_ctx l n

(* When we introduce a type var, all variables of the type context
   must be shifted *)
let rec shift_tyvar_ctx ctx n = match ctx with
    []          -> []
  | (v, b) :: l -> (shift_var n v, b) :: shift_tyvar_ctx l n

let rec varctx_ty_subst ctx ty_sub = match ctx with
    []                 -> []
  | (v, (si, ty)) :: l -> (v, (si, ty_sub ty)) :: varctx_ty_subst l ty_sub

let ctx_ty_subst ctx ty_sub =
    {ctx with var_ctx = varctx_ty_subst ctx.var_ctx ty_sub ;}

let rec varctx_efv ctx =
    List.fold_right (fun (_, (_, ty)) -> fun set -> VarSet.union set (ty_efv ty)) 
      ctx VarSet.empty

let ctx_efv ctx =
    varctx_efv ctx.var_ctx

(* Extend the context with a new variable binding. We just shift term variables *)
(* Really not needed, the variable index is just for debug purposes *)
let extend_var id bi ctx =
  let n_var = { dvi with
    v_name  = id;
    v_type  = BiVar;
    v_index = 1;
    v_size  = (length ctx.var_ctx) + 1;
  } in
  let s_ctx = shift_term_var_ctx ctx.var_ctx 1 in
  {
    var_ctx   = (n_var, bi) :: s_ctx;
    tyvar_ctx = ctx.tyvar_ctx ;
  }

(* Extend the context with a new binding. Return the new variable and
   the new context. *)
let extend_tyvar id bi ctx =
  let n_var = { dvi with
    v_name = id;
    v_type = BiTyVar;
    v_index = 1;
    v_size = (length ctx.tyvar_ctx) + 1;
  } in
  let s_tyctx = shift_tyvar_ctx    ctx.tyvar_ctx 1 in
  let s_tmctx = shift_type_var_ctx ctx.var_ctx   1 in
  {
    var_ctx   = s_tmctx ;
    tyvar_ctx = (n_var, bi) :: s_tyctx ;
  }

let remove_first_var ctx =
  if ctx.var_ctx = [] then
  (* We got no generated context, we are coming from a constant term *)
    ctx
  else
    let s_ctx = shift_term_var_ctx ctx.var_ctx (-1) in
    {
      var_ctx   = List.tl s_ctx;
      tyvar_ctx = ctx.tyvar_ctx ;
    }

(* Accessing to the variable in the context *)
let access_var ctx i = List.nth ctx.var_ctx (i-1)
let access_ty_var ctx i = List.nth ctx.tyvar_ctx (i-1)

