open Unix

open Ctx
open Syntax
open Backend
open Typing

open Support.Options
open Support.Error

let outfile    = ref (None : string option)
let infile     = ref ("" : string)

let argDefs = [
  "-o",                Arg.String (fun s -> outfile := Some s),        "File to write the generated code to" ;

  "-v",                Arg.Int  (fun l  -> debug_options := {!debug_options with level = l} ),       "Set printing level to n (1 Warning [2 Info], 3+ Debug)" ;
  "--verbose",         Arg.Int  (fun l  -> debug_options := {!debug_options with level = l} ),       "Set printing level to n (1 Warning [2 Info], 3+ Debug)" ;

  "--disable-types",   Arg.Unit (fun () -> comp_disable TypeChecker ), "Disable type checking and inference" ;
  "--disable-codegen", Arg.Unit (fun () -> comp_disable Backend     ), "Disable code generation" ;

  "--disable-unicode", Arg.Unit (fun () -> debug_options := {!debug_options with unicode = false} ), "Disable unicode printing" ;
  "--enable-annot",    Arg.Unit (fun () -> debug_options := {!debug_options with pr_ann  = true} ),  "Enable printing of type annotations" ;
  "--print-var-full",  Arg.Unit (fun () -> debug_options := {!debug_options with var_output = PrVarBoth} ), "Print names and indexes of variables" ;
  "--print-var-index", Arg.Unit (fun () -> debug_options := {!debug_options with var_output = PrVarIndex}), "Print just indexes of variables" ;
]

let dp = Support.FileInfo.dummyinfo
let main_error = error_msg General

let main_warning fi = message 1 General fi
let main_info    fi = message 2 General fi
let main_info2   fi = message 3 General fi
let main_debug   fi = message 4 General fi

let parseArgs () =
  let inFile = ref (None : string option) in
  Arg.parse argDefs
     (fun s ->
       match !inFile with
         Some(_) -> main_error dp "You must specify exactly one input file"
       | None -> inFile := Some(s))
     "Usage: fuzz [options] inputfile";
  match !inFile with
      None    -> main_error dp "No input file specified (use --help to display usage info)"; ""
    | Some(s) -> s

(* Parse the input *)
let parse file =
  let readme,writeme = Unix.pipe () in
  ignore (Unix.create_process
      "/usr/bin/cpp" [|"/usr/bin/cpp" ; "-w" ; "-include" ; "lib/primitives.fz" ; file |]
      Unix.stdin writeme Unix.stderr);
  Unix.close writeme;
  let pi = Unix.in_channel_of_descr readme in
  let lexbuf = Lexer.create file pi in
  let program =
    try Parser.body Lexer.main lexbuf
    with Parsing.Parse_error -> error_msg Parser (Lexer.info lexbuf) "Parse error" in
  Parsing.clear_parser();
  close_in pi;
  program

let call_cs name use file =
  (* main_warning dp "@[Please run the solver %s: $ %s%s@]" name use file *)
  main_info dp "@[Calling the solver %s: $ %s%s@]" name use file;
  let command = use ^ file           in
  let _z3_exit = Sys.command command in
  ()

let output_cs name file use pp cs =
  let out       = open_out file                       in
  let ofmt      = Format.formatter_of_out_channel out in
  pp ofmt cs;
  close_out out;
  call_cs name use file

let rec process_cs cs l = match l with
    [] -> ()
  | (name, file, use, pp) :: l ->
    output_cs name file use pp cs;
    process_cs cs l

let clean_pyname name =
  let pname = Filename.chop_extension name in
  let cname = Util.map (fun c -> if c = '-' then '_' else c) pname in
  cname ^ ".py"

let type_check program =
  let (cs, (ty, _ctx)) = get_type program  in
  let (di, cl) = Constraint.clean_simpl cs in

  main_info  dp "Type of the program: @[%a@]" Print.pp_type ty;
  main_info2 dp "Constraint store [full]: @[%a@]@\n@[<v>%a@]"
    Constraint.pp_cs_dirty di Constraint.pp_si_store cl;
  main_info2 dp "Constraint store [eq reduced]:@\n@[%a@]@\n@[<v>%a@]"
    Constraint.pp_cs_dirty di Constraint.pp_si_store (Constraint.simpl_subst cl);

  if comp_enabled CS then
    let z3_py_params   = ("Z3-Python", clean_pyname !infile, "python solver/z3min.py ", Z3py.gen_z3_py_constraint)   in
    (* let eclipse_params = ("ECLIPSE",   !infile ^ ".pl",      "eclipse -e fuzz -b ",     Eclipse.pp_si_store_eclipse) in *)
    (* let z3_params      = ("Z3",        !infile ^ ".z3",      "z3 -smt2 ",               Z3.gen_z3_constraint)        in *)
    (* main_warning dp "Typecheck is not complete until one of the following constraint solvers is run"; *)
    (* process_cs cs [z3_py_params; eclipse_params]; *)
    process_cs cs [z3_py_params];
  else
    main_warning dp "Constraint solving is not enabled, typechecking is not complete"

let gen_caml program outfile =
  let out  = open_out outfile in
  let ofmt = Format.formatter_of_out_channel out in
  Backend.gen_program ofmt program

(* Must use this *)
let get_tty_size = ()

(* === The main function === *)
let main () =

  (* Setup the pretty printing engine *)

  let fmt_margin =
    try
      snd (Util.get_terminal_size ())
    with _ ->
      main_warning dp "Failed to get terminal size value.";
      120
  in

  let set_pp fmt =
    Format.pp_set_ellipsis_text fmt "[...]";
    Format.pp_set_margin fmt (fmt_margin + 1); (* Don't ever ask *)
    Format.pp_set_max_indent fmt fmt_margin    in

  set_pp Format.std_formatter;
  set_pp Format.err_formatter;

  (* Read the command-line arguments *)
  infile  := parseArgs();

  let program    = parse !infile              in

  (* Print the results of the parsing phase *)
  main_debug dp "Parsed program:@\n@[%a@]@." Print.pp_term program;

  if comp_enabled TypeChecker then
    type_check program;

  if comp_enabled Backend then
    match !outfile with
    | None       -> main_warning dp "No executable was specified, use -o to generate an executable file"
    | Some out_f ->
      let out_ml = out_f ^ ".ml" in
      let command = "camlc -custom -I runtime/ runtime/pcre.o pcre.zo math.zo item.zo bag.zo init.zo " ^
	            " -lpcre " ^ out_ml ^ " -o " ^ out_f in
      gen_caml program out_ml;

      main_info dp "Compiling: %s" command;
      let _caml_exit = Sys.command command in
      ()

(* This is the way to go in order to solve the -L and pcre problem *)
(* In gneeral, we'd need this strategy to implement linking of fuzz
   programs to any arbirtrary system library *)

(* if $linkalso && test -n "$linkfiles"; then *)
(*   camlrun $stdlib/camllink -stdlib $stdlib $includes $custom $linkopt \ *)
(*     -exec $linkout $stdlib/stdlib$p.zo $linkfiles $prodump || exit $? *)
(*   if test -n "$custom"; then *)
(*     if mv $linkout /tmp/camlcode.$$ \ *)
(*        && $cc -I$stdlib -o $linkout $ccopt /tmp/camlprim.$$.c $ccfiles  \ *)
(*               -L$stdlib $cclib -lcaml  -lm \ *)
(*        && strip $linkout \ *)
(*        && cat /tmp/camlcode.$$ >> $linkout *)
(*     then *)
(*       rm -f /tmp/camlprim.$$.c camlprim.$$.o /tmp/camlcode.$$; exit 0 *)
(*     else *)
(*       rm -f $linkout /tmp/camlprim.$$.c camlprim.$$.o /tmp/camlcode.$$; exit 2 *)
(*     fi *)

(* === Call the main function and catch any exceptions === *)

let res =
  try main();
      0
  with Exit x -> x
let () = exit res
