open Format
open Support.FileInfo

(* ---------------------------------------------------------------------- *)
(* Abstract Syntax Tree for sensitivities, terms and types                *)
(* ---------------------------------------------------------------------- *)

(* TODO: Modularize. TyLollipop -> Type.Lollipop, etc... *)

(* Binders are represented using Debruijn notation *)

(* Different types of variable binding, for debug purposes *)
type fuzz_binding =
    BiVar    (* Regular varible *)
  | BiTyVar  (* Type variable   *)
  | BiETyVar (* Existential variable  *)

(* To keep things simple, we include some meta-information about
   variables that in spirit should go in each particular case, for
   instance v_interesting make only sense for SiEvars ....
*)
type var_info = {
  v_index : int;

  (* Debug fields *)
  v_name  : string;
  v_type  : fuzz_binding;
  v_size  : int;

  (* Information for existential variables *)
  v_interesting : bool; (* This is a user specified variable *)
  v_minimize    : bool; (* This is a user specified variable *)
}

(* Default varinfo *)
let dvi = {
  v_index       = -1;
  v_name        = "deadbeef";
  v_type        = BiVar;
  v_size        = -1;
  v_interesting = false;
  v_minimize    = false;
}

(* By default, the name is printed on screen, but internally it is
   ignored, the index field will always take precedence. See the
   compiler options to enable printing of indexes.
*)

(* Helper to modify the index *)
let shift_var n v = { v with
  v_index = v.v_index + n ;
  v_size  = v.v_size  + n;}

(* Set of variables *)
module OrderedVar = struct
  type t = var_info

  let compare v1 v2 = v1.v_index - v2.v_index
end

module VarSet = Set.Make(OrderedVar)

(* All of the fields are debug information *)
type binder_info = {
  b_name : string;
  b_type : fuzz_binding;
  b_size : int;          (* How many outside binders we had when this binded was found *)
}

(* Kinds for type variables *)
type kind =
    Star
  | Size
  | Sens
  | SensSchema

(* Terminology:

   We have two types of variables, existential (or unification
   variables) and universal ones. We use DeBruijn notation for the
   first ones, regular integers for the second.

   subst  operations operate on universal variables.
   esubst operations operate on existential variables.

   Every mapping operation on terms takes a parameter stating how many
   binders were traversed when the operation is called. This is
   necessary in order to properly implement most useful operations.
*)

(* Sizes and Sensitivities *)

(* This datatype is really a part of the datype type, in fact, SiVar
   are binder in TySiAbs *)
type si =
    SiVar      of var_info        (* Only used in the dependent case *)
  | SiEVar     of var_info
  | SiInfty    (* ∞ *)
  | SiIntConst of int
  | SiConst    of float
  | SiAdd      of si * si
  | SiMult     of si * si

(* Map over the variables of a sensitivity type *)
let rec si_map n f g si = match si with
    SiVar(v)      -> f n v
  | SiEVar(v)     -> g n v
  | SiInfty       -> SiInfty
  | SiIntConst(i) -> SiIntConst(i)
  | SiConst(f)    -> SiConst(f)
  | SiAdd(x, y)   -> SiAdd (si_map n f g x, si_map n f g y)
  | SiMult(x, y)  -> SiMult(si_map n f g x, si_map n f g y)

(* XXX: The shifting used here works by chance, a correct
   implementation should be named shift_over ...
*)

(* Operation that adds all the variables *)
let si_shift n si =
  let f   _ v  = SiVar (shift_var n v) in
  let eid _ v  = SiEVar v              in
  si_map n f eid si

(* Substitution si[t/x] for vars and evars *)
(* Not correct *)
let si_subst x t si =
  let sub n v = if x = v.v_index then t n else SiVar v in
  let eid _ v = SiEVar v                               in
  si_map sub eid si

let si_esubst m x t si =
  let vid _ v = SiVar v                                 in
  let sub n v = if x = v.v_index then t n else SiEVar v in
  si_map m vid sub si

(* Free variables of a term. *)
(* Use an acummulator? *)
let rec si_efv si = match si with
    SiVar(_)      -> VarSet.empty
  | SiEVar(v)     -> VarSet.add v VarSet.empty
  | SiAdd(x, y)   -> VarSet.union (si_efv x) (si_efv y)
  | SiMult(x, y)  -> VarSet.union (si_efv x) (si_efv y)
  | _             -> VarSet.empty

(* Generation of fresh sensitivity variables *)
let vsi_fresh =
  let counter = ref 0 in
  fun () ->
    let f = !counter in
    incr counter;
    { dvi with
      v_name = "S" ^ (string_of_int f);
      v_type = BiETyVar;
      v_index = f;
    }

let si_fresh () = SiEVar (vsi_fresh ())

(* Get a fresh si variable if not bound, like Option.map *)
let si_fresh_nb osi =
  match osi with
  | None    -> si_fresh ()
  | Some si -> si

(* Types *)

(* Primitive types *)
type ty_prim =
    PrimNum
  | PrimInt
  | PrimUnit
  | PrimBool
  | PrimString
  | PrimClipped

(* Types with one argument *)
(* XXX: Extend to types with n-ary arguments *)
type ty_prim1 =
    Prim1Set
  | Prim1Bag
  | Prim1Fuzzy

(* Strings in the binders just for debug purposes *)
type ty =
  (* variables used in bindings *)
    TyVar  of var_info

  (* Mainly for type checking and instances of type schemes *)
  | TyEVar of var_info

  (* Primitive types *)
  | TyPrim  of ty_prim
  | TyPrim1 of (ty_prim1 * ty)

  (* ADT *)
  | TyUnion     of ty * ty
  | TyTensor    of ty * ty
  | TyAmpersand of ty * ty

  (* Functional type *)
  | TyLollipop of ty * si * ty

  (* Fixpoint type *)
  | TyMu of binder_info * ty

  (* Quantified types *)
  | TyForall of binder_info * kind * ty
  | TyExistsSize of binder_info * ty

  (********************************************************************)
  (* Dependent types *)
  | TySizedNat of si
  | TySizedNum of si
  | TyList     of ty * si


(* map over types, first argument: action on vars, second argument
   action on evars, third argument action on sensitivities *)
let rec ty_map_aux n fv fev fsi ty = match ty with
    TyVar v                 -> fv  n v
  | TyEVar v                -> fev n v
  | TyPrim tp               -> TyPrim tp
  | TyPrim1 (tp, ty)        -> TyPrim1 (tp, ty_map_aux n fv fev fsi ty)
  (* Dependent types *)
  | TySizedNat si           -> TySizedNat(fsi n si)
  | TySizedNum si           -> TySizedNum(fsi n si)
  | TyList (ty, si)         -> TyList(ty_map_aux n fv fev fsi ty, fsi n si)
  (* ADT *)
  | TyUnion(ty1, ty2)       -> TyUnion    (ty_map_aux n fv fev fsi ty1, ty_map_aux n fv fev fsi ty2)
  | TyTensor(ty1, ty2)      -> TyTensor   (ty_map_aux n fv fev fsi ty1, ty_map_aux n fv fev fsi ty2)
  | TyAmpersand(ty1, ty2)   -> TyAmpersand(ty_map_aux n fv fev fsi ty1, ty_map_aux n fv fev fsi ty2)
  (* *)
  | TyLollipop(ty1, s, ty2) -> TyLollipop(ty_map_aux n fv fev fsi ty1, fsi n s, ty_map_aux n fv fev fsi ty2)

  | TyMu(b, ty)             -> TyMu(b, ty_map_aux (n+1) fv fev fsi ty)
  (* *)
  | TyForall(b, k, ty)      -> TyForall(b, k, ty_map_aux (n+1) fv fev fsi ty)
  | TyExistsSize(b, ty)     -> TyExistsSize(b, ty_map_aux (n+1) fv fev fsi ty)

let ty_map fv fev fsi ty = ty_map_aux 0 fv fev fsi ty

(* Generating fresh type variables *)
let ty_fresh =
  let counter = ref 0 in
  fun () ->
    let f = !counter in
    incr counter;
    TyEVar { dvi with
      v_name = "T" ^ (string_of_int f);
      v_type = BiETyVar;
      v_index = f;
    }

(* XXX: The shifting used here works by chance, a correct
   implementation should be named shift_over ...
*)
let shift_type n ty =
  let fv  _ v = TyVar (shift_var n v) in
  let fev _ v = TyEVar v              in
  let fsi _   = si_shift n            in
  ty_map fv fev fsi ty

(* DeBruijn substitution ty[t/x] only for variables binded by
   TyForall! that is to say, of kind Star *)
let ty_subst x t ty =
  let fv  n  v =
    (* Is this the variable we are looking for? *)
    if (x+n) = v.v_index then
      t n
    (* If the variable has index greater than n, decrement by one *)
    else if v.v_index > (x+n) then
      TyVar (shift_var (-1) v)
    (* Nothing to do, captured variable *)
    else
      TyVar {v with v_size = v.v_size - 1;}
  in
  let fev _  v = TyEVar v             in
  (* Sensitivities also have type variables inside, however we need a
     different function for them ...*)
  let fsi _ si = si                   in
  ty_map fv fev fsi ty

(* EVar substitution *)
let ty_esubst x t ty =
  let fv  _ v  = TyVar v                                 in
  let fev n v  = if x = v.v_index then t n else TyEVar v in
  let fsi _ si = si                                      in
  ty_map fv fev fsi ty

(* List of substitutions *)

(* We could represent subs as functions from types to types, but this
   is better for debug purposes *)

(* type ty_esub = (var_info * ty) list *)

(* let rec apply_esub s ty = match s with *)
(*     []             -> ty *)
(*   | ((v, t) :: sl) -> apply_esub sl (ty_esubst v.v_index (fun _ -> t) ty) *)

(* let ty_new_esub v ty = [(v, ty)] *)

(* let ty_esub_id         = [] *)
(* let ty_esub_comp s1 s2 = s1 @ s2 *)


(* let apply_esub s ty = *)
(*   let fv   = ty_efv ty                                           in *)
(*   let fv_s = List.fold_left *)
(*     (fun fv_se -> fun (_, ty_s) -> VarSet.union (ty_efv ty_s) fv_se) *)
(*     VarSet.empty s                                               in *)
(*   let fv_all = VarSet.union fv fv_s                              in *)
(*   let s'     = List.filter (fun (v, _) -> VarSet.mem v fv_all) s in *)
(*   apply_esub s' ty *)

(* Quite a bit faster *)
type ty_esub = ty -> ty

let apply_esub s ty = s ty

let ty_new_esub v ty   = ty_esubst v.v_index (fun _ -> ty)
let ty_esub_id  ty     = ty
let ty_esub_comp s1 s2 = fun ty -> s2 (s1 ty)

(* Closes binder under set *)
let fv_binder s =
  (* Remove variable one *)
  let s' = VarSet.filter (fun var -> var.v_index <> 1) s in
  (* Shift the rest of variables *)
  VarSet.fold (fun var -> fun set -> VarSet.add (shift_var (-1) var) set) s' VarSet.empty

(* Could be done with a fold *)
let rec ty_fv ty = match ty with
    TyVar v                  -> VarSet.add v VarSet.empty
  | TyEVar _v                -> VarSet.empty
  | TyPrim _tp               -> VarSet.empty
  | TyPrim1(_tp, ty)         -> ty_fv ty

  (* Dependent types *)
  | TySizedNat _si           -> VarSet.empty
  | TySizedNum _si           -> VarSet.empty
  | TyList(ty, _si)          -> ty_fv ty

  (* ADT *)
  | TyUnion(ty1, ty2)       -> VarSet.union (ty_fv ty1) (ty_fv ty2)
  | TyTensor(ty1, ty2)      -> VarSet.union (ty_fv ty1) (ty_fv ty2)
  | TyAmpersand(ty1, ty2)   -> VarSet.union (ty_fv ty1) (ty_fv ty2)

  (* *)
  | TyLollipop(ty1, _s, ty2) -> VarSet.union (ty_fv ty1) (ty_fv ty2)
  | TyMu(_b, ty)             -> fv_binder (ty_fv ty)
  (* *)
  | TyForall(_b, _k, ty)     -> fv_binder (ty_fv ty)
  | TyExistsSize(_b, ty)     -> fv_binder (ty_fv ty)

(* Set of free existential vars *)
let rec ty_efv ty = match ty with
    TyVar _v                 -> VarSet.empty
  | TyEVar v                 -> VarSet.add v VarSet.empty
  | TyPrim _tp               -> VarSet.empty
  | TyPrim1(_tp, ty)         -> ty_efv ty

  (* Dependent types *)
  | TySizedNat _si           -> VarSet.empty
  | TySizedNum _si           -> VarSet.empty
  | TyList(ty, _si)          -> ty_efv ty

  (* ADT *)
  | TyUnion(ty1, ty2)       -> VarSet.union (ty_efv ty1) (ty_efv ty2)
  | TyTensor(ty1, ty2)      -> VarSet.union (ty_efv ty1) (ty_efv ty2)
  | TyAmpersand(ty1, ty2)   -> VarSet.union (ty_efv ty1) (ty_efv ty2)

  (* *)
  | TyLollipop(ty1, _s, ty2) -> VarSet.union (ty_efv ty1) (ty_efv ty2)
  | TyMu(_b, ty)             -> (ty_efv ty)
  (* *)
  | TyForall(_b, _k, ty)     -> (ty_efv ty)
  | TyExistsSize(_b, ty)     -> (ty_efv ty)

let ty_unfold ty = match ty with
  | TyMu(b, ty_i) -> ty_subst 1 (fun _ -> TyMu(b, ty_i)) ty_i
  | _             -> ty


(* The fuzz language can be reduced to the following core:

  - var
  - const
  - abstraction
  - application
  - ampersand

   We only need let to introduce polymorphic definitions, and
   fold/unfold to support recursive data types.
*)

(*********************************************************************)
(* Terms                                                             *)

(* Primitive Terms *)
type term_prim =
    PrimTUnit
  | PrimTNum    of float
  | PrimTInt    of int
  | PrimTBool   of bool
  | PrimTString of string
  | PrimTFun    of string * ty

let type_of_prim t = match t with
    PrimTUnit       -> TyPrim PrimUnit
  | PrimTNum _      -> TyPrim PrimNum
  | PrimTInt _      -> TyPrim PrimInt
  | PrimTBool _     -> TyPrim PrimBool
  | PrimTString  _  -> TyPrim PrimString
  | PrimTFun(_, ty) -> ty

type term =
    TmVar of info * var_info

  (* Will remove once we have sensitivity abstraction *)
  | TmTensDest  of info * binder_info * binder_info * term * term
  | TmUnionCase of info * term * binder_info * term * binder_info * term
  (*                      t of { inl(x)     => tm1  | inl(y)     => tm2  } *)

  (* Primitive terms *)
  | TmPrim     of info * term_prim

  (* Primitive adaptable noise function *)
  | TmFuzz     of info * float

  (* The three fundamental constructs of our language: *)

  (* Regular Abstraction and Applicacion *)
  | TmApp of info * term * term

  (* In a lambda is possible to independently annotate the input and return type *)
  | TmAbs of info * binder_info * (si * ty) option * ty option * term

  (* & constructor *)
  | TmAmpersand of info * term * term

  (* Recursive data types *)
  | TmFold    of info * ty * term
  | TmUnfold  of info * term

  (* Only needed for polymorphism *)
  | TmLet      of info * binder_info * ty option * term * term

  (***********************************************************************)
  (* What to do with this, what is the typing rule?                      *)
  | TmTypedef of info * binder_info * ty * term

  (***********************************************************************)
  (* Dependent case expressions, removal pending on type constraints     *)

  (*                      t of {nil => tm1  | (    x     ::     xs      ) [si]       => tm2 } *)
  | TmListCase  of info * term       * term * binder_info * binder_info * binder_info * term
  (*                      t of {Z => tm1  | (S x)         [si]       => tm2 }                 *)
  | TmNatCase   of info * term     * term * binder_info * binder_info * term

  (* Pack/Unpack *)
  | TmPack of info * term * si * ty
  | TmUnpack of info * binder_info * binder_info * term * term

  (* Sensitity Abstraction and Applicacion *)
  | TmSiAbs of info * binder_info * kind * term
  | TmSiApp of info * term * si

let map_prim_ty n f p =
  match p with
    PrimTUnit       -> p
  | PrimTNum(_)     -> p
  | PrimTInt(_)     -> p
  | PrimTBool(_)    -> p
  | PrimTString(_)  -> p
  | PrimTFun(s, ty) -> PrimTFun(s, f n ty)

(* XXX: Extend map to sensitivities *)
let rec map_term_ty_aux n f tm =
  let tf n = map_term_ty_aux n f                       in
  let opf  = Option.map (f n)                          in
  let opsf = Option.map (fun (si, ty) -> (si, f n ty)) in
  match tm with
    TmVar(i, v)                -> TmVar (i, v)
  | TmFuzz(i, n)               -> TmFuzz (i, n)
  | TmPrim(i, p)               -> TmPrim(i, map_prim_ty n f p)

  (* Will die soon *)
  | TmTensDest(i, bi_x, bi_y,      tm,      tm_i) ->
    TmTensDest(i, bi_x, bi_y, tf n tm, tf n tm_i)

  | TmUnionCase(i,      tm, bi_l,      tm_l, bi_r,      tm_r)  ->
    TmUnionCase(i, tf n tm, bi_l, tf n tm_l, bi_r, tf n tm_r)

  (* The three fundamental constructs of the language *)
  | TmAbs(i, bi,      osity,     orty,      tm)           ->
    TmAbs(i, bi, opsf osity, opf orty, tf n tm)

  | TmApp(i,      tm1,      tm2)          ->
    TmApp(i, tf n tm1, tf n tm2)

  (* *)
  | TmAmpersand(i,      tm1,      tm2)    ->
    TmAmpersand(i, tf n tm1, tf n tm2)

  (* Only needed for polymorphism *)
  | TmLet(i, bi,     oty,      tm,      tm_i) ->
    TmLet(i, bi, opf oty, tf n tm, tf n tm_i)

  (* Recursive datatypes *)
  | TmFold(i,     ty,      tm)           ->
    TmFold(i, f n ty, tf n tm)

  | TmUnfold(i,      tm) ->
    TmUnfold(i, tf n tm)

  (* Type Binder! *)
  | TmTypedef(i, bi,     ty,          tm)        ->
    TmTypedef(i, bi, f n ty, tf (n+1) tm)

  (* Dependent stuff *)
  | TmListCase(i,      tm,      tm_l, bi_x, bi_xs, bi_si,      tm_r) ->
    TmListCase(i, tf n tm, tf n tm_l, bi_x, bi_xs, bi_si, tf n tm_r)

  | TmNatCase(i,      tm,      tm_l, bi_n, bi_si,      tm_r)    ->
    TmNatCase(i, tf n tm, tf n tm_l, bi_n, bi_si, tf n tm_r)

  (* Sensitivities *)
  | TmSiApp (i,      tm, si)         ->
    TmSiApp (i, tf n tm, si)

  (* Type binder! *)
  | TmSiAbs (i, bi, k,          tm)  ->
    TmSiAbs (i, bi, k, tf (n+1) tm)

  (* XXX: TODO: one is a binder *)
  | TmPack (i,      tm, si,     ty)      ->
    TmPack (i, tf n tm, si, f n ty)

  | TmUnpack (i, bi, bi',      tm1,      tm2)  ->
    TmUnpack (i, bi, bi', tf n tm1, tf n tm2)

let map_term_ty f tm = map_term_ty_aux 0 f tm

(* tm[t/x] *)
let term_ty_subst x t tm =
  let sub n = ty_subst (n+x) t in
  map_term_ty sub tm

(************************************************************************)
(* File info extraction *)

let tmInfo t = match t with
    TmVar(fi, _)               -> fi
  | TmFuzz(fi, _)              -> fi
  | TmPrim(fi, _)              -> fi

  (* Will die soon *)
  | TmTensDest(fi,_,_,_,_)     -> fi
  | TmUnionCase(fi,_,_,_,_,_)  -> fi

  (* The three fundamental constructs of the language *)
  | TmAbs(fi,_,_,_,_)        -> fi
  | TmApp(fi, _, _)            -> fi
  (* *)
  | TmAmpersand(fi,_,_)        -> fi

  (* Only needed for polymorphism *)
  | TmLet(fi,_,_,_,_)          -> fi

  (* Recursive datatypes *)
  | TmFold(fi,_,_)             -> fi
  | TmUnfold(fi,_)             -> fi

  (* Dependent stuff *)
  | TmListCase(fi,_,_,_,_,_,_) -> fi
  | TmNatCase(fi,_,_,_,_,_)    -> fi

  (* *)
  | TmSiApp (fi, _, _)         -> fi
  | TmSiAbs (fi, _, _, _)      -> fi

  | TmUnpack (fi, _, _, _, _)  -> fi
  | TmPack (fi, _, _, _)       -> fi

  (* Misc *)
  | TmTypedef(fi,_,_,_)        -> fi
