open Support.FileInfo

(* TODO: Check what can we make private *)

(* Different types of variable binding, for debug purposes *)
type fuzz_binding =
    BiVar    (* Regular varible *)
  | BiTyVar  (* Type variable   *)
  | BiETyVar (* Existential variable  *)

(* Variables and binders *)
type var_info = {
  v_index : int;

  (* Debug fields *)
  v_name  : string;
  v_type  : fuzz_binding;
  v_size  : int;

  (* Information for existential variables *)
  v_interesting : bool; (* This is a user specified variable *)
  v_minimize    : bool; (* This is a user specified variable *)
}

(* Default var_info *)
val dvi : var_info

(* Adds n to a var_info *)
val shift_var : int -> var_info -> var_info

(* Sets of variables *)
module VarSet : Set.S with type elt = var_info

(* All of the fields are debug information *)
type binder_info = {
  b_name : string;
  b_type : fuzz_binding;
  b_size : int;          (* How many outside binders we had when this binded was found *)
}

(* Types *)

(* Part 0: kinds *)
type kind =
    Star
  | Size
  | Sens
  | SensSchema

(* Part 1: Sensitivites *)
type si =
    SiVar      of var_info        (* Only used in the dependent case *)
  | SiEVar     of var_info
  | SiInfty    (* ∞ *)
  | SiIntConst of int
  | SiConst    of float
  | SiAdd      of si * si
  | SiMult     of si * si


(* Shift variable indexes by n *)
val si_shift : int -> si -> si

val vsi_fresh   : unit -> var_info
val si_fresh    : unit -> si
val si_fresh_nb : si option -> si

val si_efv : si -> VarSet.t

(* Number of binders, index, and sens *)
val si_esubst : int -> int -> (int -> si) -> si -> si

(* Part 2: Regular "HM" types *)
(* Primitive types *)
type ty_prim =
    PrimNum
  | PrimInt
  | PrimUnit
  | PrimBool
  | PrimString
  | PrimClipped

(* Types with one argument *)
(* XXX: Extend to types with n-ary arguments *)
type ty_prim1 =
    Prim1Set
  | Prim1Bag
  | Prim1Fuzzy

type ty =
  (* variables used in bindings *)
    TyVar  of var_info

  (* Mainly for type checking and instances of type schemes *)
  | TyEVar of var_info

  (* Primitive types *)
  | TyPrim  of ty_prim
  | TyPrim1 of (ty_prim1 * ty)

  (* ADT *)
  | TyUnion     of ty * ty
  | TyTensor    of ty * ty
  | TyAmpersand of ty * ty

  (* Functional type *)
  | TyLollipop of ty * si * ty

  (* Fixpoint type *)
  | TyMu of binder_info * ty

  (* Quantified types *)
  | TyForall of binder_info * kind * ty
  | TyExistsSize of binder_info * ty

  (********************************************************************)
  (* Dependent types *)
  | TySizedNat of si
  | TySizedNum of si
  | TyList     of ty * si

(* XXX: This is incorrect, right now it shifts all the indexes, thus
   it is buggy, see the other comment *)

(* Shift all the open indexes by n *)
val shift_type : int -> ty -> ty

val ty_fresh   : unit -> ty

(* Free Existential variables of a type *)
val ty_efv : ty -> VarSet.t

(* Substitutions *)

(* Capture avoiding sub, the term must be dependent on the number of
   binders under it is replaced *)
val ty_subst  : int -> (int -> ty) -> ty -> ty

(* Same for existential variables *)
val ty_esubst : int -> (int -> ty) -> ty -> ty

(* Unfolding of types: One step *)
val ty_unfold    : ty -> ty

(* Substitutions lists *)
(* XXX: This is horrible and must improve *)
type ty_esub

val ty_esub_id   : ty_esub
val ty_new_esub  : var_info -> ty -> ty_esub

val apply_esub   : ty_esub -> ty -> ty
val ty_esub_comp : ty_esub -> ty_esub -> ty_esub

(* Terms *)

(* Primitive Terms *)
type term_prim =
    PrimTUnit
  | PrimTNum    of float
  | PrimTInt    of int
  | PrimTBool   of bool
  | PrimTString of string
  | PrimTFun    of string * ty

val type_of_prim : term_prim -> ty

type term =
    TmVar of info * var_info

  (* Will remove once we have sensitivity abstraction *)
  | TmTensDest  of info * binder_info * binder_info * term * term
  | TmUnionCase of info * term * binder_info * term * binder_info * term
  (*                      t of { inl(x)     => tm1  | inl(y)     => tm2  } *)

  (* Primitive terms *)
  | TmPrim     of info * term_prim

  (* Primitive adaptable noise function *)
  | TmFuzz     of info * float

  (* The three fundamental constructs of our language: *)

  (* Regular Abstraction and Applicacion *)
  | TmApp of info * term * term

  (* In a lambda is possible to independently annotate the input and return type *)
  | TmAbs of info * binder_info * (si * ty) option * ty option * term

  (* & constructor *)
  | TmAmpersand of info * term * term

  (* Recursive data types *)
  | TmFold    of info * ty * term
  | TmUnfold  of info * term

  (* Only needed for polymorphism *)
  | TmLet      of info * binder_info * ty option * term * term

  (***********************************************************************)
  (* What to do with this, what is the typing rule?                      *)
  | TmTypedef of info * binder_info * ty * term

  (***********************************************************************)
  (* Dependent case expressions, removal pending on type constraints     *)

  (*                      t of {nil => tm1  | (    x     ::     xs      ) [si]       => tm2 } *)
  | TmListCase  of info * term       * term * binder_info * binder_info * binder_info * term
  (*                      t of {Z => tm1  | (S x)         [si]       => tm2 }                 *)
  | TmNatCase   of info * term     * term * binder_info * binder_info * term

  (* Pack/Unpack *)
  | TmPack of info * term * si * ty
  | TmUnpack of info * binder_info * binder_info * term * term

  (* Sensitity Abstraction and Applicacion *)
  | TmSiAbs of info * binder_info * kind * term
  | TmSiApp of info * term * si

val tmInfo : term -> info

(* Substitution for type annotations *)
(* tm[t/x] *)
val term_ty_subst : int -> (int -> ty) -> term -> term

