open Ctx
open Syntax
open Constraint

open Support.Error
open Support.FileInfo

(* Todo: the i parameter could be made part of the monad *)

(* The type engine is composed of four components:

   - A type monad that keeps state, as of now a constraint store.

   - A type constraint engine that checks equalities of types and
     interfaces with the external solvers.

   - A context management module.

     Every typing computation happens in a context, but the flow is
     too complex at the moment to try represent it a state of the
     typer.

   - The implementation of the typing rules itself, by program
     travesal.

   Note on the substitution strategy: We keep a substitution, but we
   must be sure that is lazily applied (with normalize_* ) whenever we
   are going to compare types.

*)

(* Errors *)
type ty_error_elem =
| SensError    of si_cstr
| MoreGeneral  of ty * ty
| TypeMismatch of ty * ty
| TypeInst     of ty * ty
| CannotApply  of ty * ty
| OccursCheck  of var_info * ty
| WrongShape   of ty * string

(* Not sure about this yet *)
exception TypeError of ty_error_elem withinfo

let typing_error fi = error_msg    Support.Options.TypeChecker fi
let typing_error_pp = error_msg_pp Support.Options.TypeChecker

let ty_warning fi = message 1 Support.Options.TypeChecker fi
let ty_info    fi = message 2 Support.Options.TypeChecker fi
let ty_info2   fi = message 3 Support.Options.TypeChecker fi
let ty_debug   fi = message 4 Support.Options.TypeChecker fi
let ty_debug2  fi = message 5 Support.Options.TypeChecker fi
let ty_debug3  fi = message 6 Support.Options.TypeChecker fi

type 'a ty_error =
  Right of 'a
| Left  of ty_error_elem withinfo

let new_error info e = { i = info; v = e ; }

module TypeMonad = struct

  (* Type checking monad for dfuzz: (ctx, cs) -> (ctx, cs, ty) option *)

  (* If needed, we could keep the state and push the error inside *)
  type 'a tyState =
    State of ((Constraint.c_store * ty_esub) -> ((Constraint.c_store * ty_esub) * 'a) ty_error)

  (* Native @@ is already in ocaml 4.0 *)
  let (@@) x y = x y

  let fail e    = State (fun _ -> Left e)
  let return x  = State (fun s -> Right (s, x))

  let (>>=) m f = State (fun s ->
    let (State mf) = m in
    match mf s with
      Left e               -> Left e
    | Right (s', a)        ->
      let State ff = f a in
      ff s'
  )

  let (>>) m f = m >>= fun _ -> f

  let get = State (fun s ->
    Right (s, s))

  let put s = State (fun _ ->
    Right (s, ()))

  let get_cs = State (fun (c, s) ->
    Right ((c, s), c))

  let put_cs c = State (fun (_, s) ->
    Right ((c, s), ()))

  let get_sub = State (fun (c, s) ->
    Right ((c, s), s))

  let put_sub s = State (fun (c, _) ->
    Right ((c, s), ()))

  let add_sub s =
    get_sub            >>= fun sub ->
    put_sub @@ ty_esub_comp sub s

  let normalize_type ty =
    get_sub            >>= fun sub ->
    return @@ apply_esub sub ty

  let normalize_ctx ctx =
    get_sub            >>= fun sub ->
    return @@ ctx_ty_subst ctx (apply_esub sub)

  let normalize (ty, ctx) =
    normalize_type ty     >>= fun ty'  ->
    normalize_ctx  ctx    >>= fun ctx' ->
    return (ty', ctx')

  let iwhen b s e = if b then return s else fail e
end

module TypeConstraint = struct
  open TypeMonad

  (* Note the repeated pattern here *)
  let post_cs i cs =
    (ty_debug i "*** Posting constraint %a" Constraint.pp_si_cstr_elem cs;
     get_cs)                                       >>= fun cstore ->
      match Constraint.post_cs cstore cs with
	None         -> fail (new_error i (SensError cs))
      | Some cstore' -> put_cs cstore'

  let post_si_eq i si1 si2 =
    (* Optimization to avoid being spammed by ctx merging *)
    if si1 = si2 then
      return ()
    else
      post_cs i (cs_eq i si1 si2)

  let post_si_ge i si1 si2 = post_cs i (cs_ge i si1 si2)

  let post_si_ge_m i si1 osi2 =
    match osi2 with
    | None     -> return ()
    | Some si2 -> post_cs i (cs_ge i si1 si2)

  (* Structural type equality *)
  (* Must be called with normalized types! *)
  let rec check_type_eq i ty_1 ty_2 =
    let ty_neq = new_error i (TypeInst(ty_1, ty_2))   in
    let iwhen b = iwhen b () ty_neq                   in

    match ty_1, ty_2 with
      TyVar v1, TyVar v2   -> iwhen (v1.v_index = v2.v_index)

    | TyEVar v1, TyEVar v2 -> iwhen (v1 = v2)

    | TyPrim p1, TyPrim p2 -> iwhen (p1 = p2)

    | TyPrim1(pl, tyl), TyPrim1(pr, tyr) ->
      iwhen (pl = pr)                                   >>
      check_type_eq i tyl tyr

    | TyUnion(tyl1, tyl2), TyUnion(tyr1, tyr2) ->
      check_type_eq i tyl1 tyr1                         >>
      check_type_eq i tyl2 tyr2

    | TyTensor(tyl1, tyl2), TyTensor(tyr1, tyr2) ->
      check_type_eq i tyl1 tyr1                         >>
      check_type_eq i tyl2 tyr2

    | TyAmpersand(tyl1, tyl2), TyAmpersand(tyr1, tyr2) ->
      check_type_eq i tyl1 tyr1                         >>
      check_type_eq i tyl2 tyr2

    | TyLollipop(tyl1, sil, tyl2), TyLollipop(tyr1, sir, tyr2) ->
      check_type_eq i tyl1 tyr1                         >>
      check_type_eq i tyl2 tyr2                         >>
      post_si_eq i sil sir

    | TyMu(_bl, tyl), TyMu(_br, tyr) ->
      check_type_eq i tyl tyr

    | TyForall(_bl, kl, tyl), TyForall(_br, kr, tyr) ->
      iwhen (kl = kr)                                   >>
      check_type_eq i tyl tyr

    | TyExistsSize(_bl, tyl), TyExistsSize(_br, tyr) ->
      check_type_eq i tyl tyr

    | TySizedNat(sil), TySizedNat(sir) ->
      post_si_eq i sil sir

    | TySizedNum(sil), TySizedNum(sir) ->
      post_si_eq i sil sir

    | TyList(tyl, sil), TyList(tyr, sir) ->
      check_type_eq i tyl tyr                           >>
      post_si_eq i sil sir

    | _, _ -> fail ty_neq

  (* Standard occurs check *)
  let check_occurs_check i index ty =
    let fv_set = ty_efv ty in
    if VarSet.mem index fv_set then
      fail @@ new_error i (OccursCheck(index, ty))
    else
      return ()

  (* ty_ann must be an instance of ty_i. Return the substitution that
     makes them equal *)
  let rec check_type_inst i ty_ann ty_i =

    normalize_type ty_ann                               >>= fun ty_ann_n ->
    normalize_type ty_i                                 >>= fun ty_i_n   ->

    let ty_neq = new_error i (TypeInst(ty_ann_n, ty_i_n))     in
    let iwhen b = iwhen b () ty_neq                           in

    match ty_ann_n, ty_i_n with
    | TyVar v1, TyVar v2   -> iwhen (v1.v_index = v2.v_index)

    (* The ann is less general than the type *)
    (* We must not perform occurs check on EVar = Evar *)
    | TyEVar _ as ty_ev, TyEVar v    ->
      add_sub (ty_new_esub v ty_ev)

    | ty_ann, TyEVar v    ->
      check_occurs_check i v ty_ann                     >>
      add_sub (ty_new_esub v ty_ann)

    | TyPrim p1, TyPrim p2 -> iwhen (p1 = p2)

    | TyPrim1(pl, tyl), TyPrim1(pr, tyr) ->
      iwhen (pl = pr)                                   >>
      check_type_inst i tyl tyr

    | TyUnion(tyl1, tyl2), TyUnion(tyr1, tyr2) ->
      check_type_inst i tyl1 tyr1                       >>
      check_type_inst i tyl2 tyr2

    | TyTensor(tyl1, tyl2), TyTensor(tyr1, tyr2) ->
      check_type_inst i tyl1 tyr1                       >>
      check_type_inst i tyl2 tyr2

    | TyAmpersand(tyl1, tyl2), TyAmpersand(tyr1, tyr2) ->
      check_type_inst i tyl1 tyr1                       >>
      check_type_inst i tyl2 tyr2

    | TyLollipop(tyl1, sil, tyl2), TyLollipop(tyr1, sir, tyr2) ->
      post_si_eq i sil sir                              >>
      check_type_inst i tyl1 tyr1                       >>
      check_type_inst i tyl2 tyr2

    | TyMu(_bl, tyl), TyMu(_br, tyr) ->
      check_type_inst i tyl tyr

    | TyForall(_bl, kl, tyl), TyForall(_br, kr, tyr) ->
      iwhen (kl = kr)                                  >>
      check_type_inst i tyl tyr

    | TyExistsSize(_bl, tyl), TyExistsSize(_br, tyr) ->
      check_type_inst i tyl tyr

    | TySizedNat(sil), TySizedNat(sir) ->
      post_si_eq i sil sir

    | TySizedNum(sil), TySizedNum(sir) ->
      post_si_eq i sil sir

    | TyList(tyl, sil), TyList(tyr, sir) ->
      post_si_eq i sil sir                            >>
      check_type_inst i tyl tyr

    | _, _ -> fail ty_neq

  (* Type unif *)
  let rec check_type_unif i ty_1 ty_2 =
    normalize_type ty_1                                 >>= fun ty1 ->
    normalize_type ty_2                                 >>= fun ty2 ->

    let ty_neq = new_error i (TypeMismatch(ty1, ty2))   in
    let iwhen b = iwhen b () ty_neq                     in

    match ty1, ty2 with

    (* Unification for evars *)
    | TyEVar v1, (TyEVar _ as ty_v) ->
      add_sub (ty_new_esub v1 ty_v)

    | TyEVar v, ty_i      ->
      check_occurs_check i v ty_i                       >>
      add_sub (ty_new_esub v ty_i)

    | ty_i, TyEVar v      ->
      check_occurs_check i v ty_i                       >>
      add_sub (ty_new_esub v ty_i)

    | TyVar v1, TyVar v2   -> iwhen (v1.v_index = v2.v_index)

    | TyPrim p1, TyPrim p2 -> iwhen (p1 = p2)

    | TyPrim1(pl, tyl), TyPrim1(pr, tyr) ->
      iwhen (pl = pr)                                   >>
      check_type_unif i tyl tyr

    | TyUnion(tyl1, tyl2), TyUnion(tyr1, tyr2) ->
      check_type_unif i tyl1 tyr1                       >>
      check_type_unif i tyl2 tyr2

    | TyTensor(tyl1, tyl2), TyTensor(tyr1, tyr2) ->
      check_type_unif i tyl1 tyr1                       >>
      check_type_unif i tyl2 tyr2

    | TyAmpersand(tyl1, tyl2), TyAmpersand(tyr1, tyr2) ->
      check_type_unif i tyl1 tyr1                       >>
      check_type_unif i tyl2 tyr2

    | TyLollipop(tyl1, sil, tyl2), TyLollipop(tyr1, sir, tyr2) ->
      post_si_eq i sil sir                              >>
      check_type_unif i tyl1 tyr1                       >>
      check_type_unif i tyl2 tyr2

    | TyMu(_bl, tyl), TyMu(_br, tyr) ->
      check_type_unif i tyl tyr

    (* This is a weird artifact of our syntax and language
       choices *)
    | TyForall(_bl, kl, tyl), TyForall(_br, kr, tyr) ->
      iwhen (kl = kr)                                  >>
      check_type_unif i tyl tyr

    | TyExistsSize(_bl, tyl), TyExistsSize(_br, tyr) ->
      check_type_unif i tyl tyr

    | TySizedNat(sil), TySizedNat(sir) ->
      post_si_eq i sil sir

    | TySizedNum(sil), TySizedNum(sir) ->
      post_si_eq i sil sir

    | TyList(tyl, sil), TyList(tyr, sir) ->
      post_si_eq i sil sir                             >>
      check_type_unif i tyl tyr

    | _, _ -> fail ty_neq

  (* Note that the type was already normalized *)
  let clipped_hack ty =
    match ty with
    | TyPrim PrimClipped -> return @@ TyPrim PrimNum
    | _                  -> return ty

  (* Checks for types of different shapes *)

  (* Check that the type is an application and can be applied to
     arg. Return the result and sensitivity of the application *)
  let check_app i ty_arr ty_arg =
    normalize_type ty_arr                                 >>= fun ty_arr ->
    normalize_type ty_arg                                 >>= fun ty_arg ->

    ty_debug i "<-> Application of @[%a@] to @[%a@]" Print.pp_type ty_arr Print.pp_type ty_arg;
    match ty_arr with
    | TyLollipop(tya, si, tyb) ->

      (* XXX: Big HACK *)
      (* If the argument is a clipped, cast it to a real *)
      clipped_hack ty_arg                                    >>= fun ty_arg ->

      (* Both types come from instantiation on the leaves, unify *)

      (* XXX: if dfuzz_enabled we ought to have subtyping here! *)
      check_type_unif i tya ty_arg                           >>
      normalize_type tyb                                     >>= fun tyb_n ->
      return @@ (tyb_n, si)

    | _                        -> fail (new_error i @@ CannotApply(ty_arr, ty_arg))

  let check_tensor_shape i ty =
    normalize_type ty    >>= fun ty ->

    match ty with
    | TyTensor(ty1, ty2) -> return (ty1, ty2)
    | _                  -> fail (new_error i @@ WrongShape (ty, "tensor"))

  let check_union_shape i ty =
    normalize_type ty    >>= fun ty ->

    match ty with
    | TyUnion(ty1, ty2) -> return (ty1, ty2)
    | _                 -> fail (new_error i @@ WrongShape (ty, "union"))

  let check_mu_shape i ty =
    normalize_type ty    >>= fun ty ->

    match ty with
    | TyMu(_, _)          -> return (ty_unfold ty)
    | _                   -> fail (new_error i @@ WrongShape (ty, "mu"))

  (* Check that the returned type is an instance of an annotation *)
  (* If true, we store the new substitution in the state          *)
  let check_type_inst_op i otya ty =
    match otya with
      None        -> return ()
    | Some ty_ann ->
      check_type_inst i ty_ann ty

  (* We consider a type scheme a forall with kind of Star *)
  (* Note that we only instantiate head variables *)
  let rec gen_type_inst ty = match ty with
      TyForall(_n, Star, ty_scheme) ->
	(* Generate a fresh evar regardless of the index *)
	let ty_f    = ty_fresh ()                          in
	let ty_inst = ty_subst 1 (fun _ -> ty_f) ty_scheme in
	gen_type_inst ty_inst
    | _                     -> return @@ ty

  (* Instantiate all the annotations of a term *)
  let rec inst_params_ann ty tm = match ty with
      TyForall(_n, Star, ty_scheme) ->
	(* Generate a fresh evar regardless of the index *)
	let ty_f    = (fun _ -> ty_fresh ())         in
	let tm_inst = term_ty_subst 1 ty_f tm        in
	let ty_inst = ty_subst      1 ty_f ty_scheme in
	inst_params_ann ty_inst tm_inst

    | _ -> tm

  let open_annot oty_ann tm = match oty_ann with
      None    -> tm
    | Some ty -> inst_params_ann ty tm

  (* generalize a type *)
  let rec create_gen_type vl ty = match vl with
      []         -> ty
    | (v :: vll) ->
      let gen_ty  = create_gen_type vll ty      in
      let n_b     = { b_name = "Gen"; b_type = BiTyVar; b_size = -1; }   in
      let n_v   n = { dvi with
	v_index = 1+n;
	v_name  = "G" ^ (string_of_int (1+n));
	v_type  = BiTyVar;
      }                                                    in
      let gen_ty' = ty_esubst v.v_index (fun n -> TyVar (n_v n)) gen_ty  in
      TyForall(n_b, Star, gen_ty')

  let generalize_type ctx ty_i =
    let ty_i_fv     = ty_efv  ty_i in
    let ctx_fv      = ctx_efv ctx  in
    let vars_to_gen = VarSet.diff ty_i_fv ctx_fv in
    let ty_g        = create_gen_type (List.rev (VarSet.elements vars_to_gen)) ty_i in
    ty_g

end

module TypeCtx = struct
  open TypeMonad
  open TypeConstraint

  let with_ctx ctx m =
    m ctx

  (* Cannot fail *)
  let with_extend_ctx ctx v ty m =
    with_ctx (extend_var v (None, ty) ctx) m      >>= fun (ty_m, ctx_m) ->

    let (_, (si_b, ty_b)) = access_var ctx_m 1    in
    return @@ ((si_b, ty_b), ty_m, remove_first_var ctx_m)

  (* Not good we have to define this... *)
  let with_extend_ctx_2 ctx v1 ty1 v2 ty2 m =
    with_ctx (extend_var v2 (None, ty2) (extend_var v1 (None, ty1) ctx)) m  >>= fun (ty, ctx_n) ->

    let (_, (si_2, _)) = access_var ctx_n 1 in
    let (_, (si_1, _)) = access_var ctx_n 2 in
    return @@ (ty, (si_1, si_2), remove_first_var (remove_first_var ctx_n))

  let rec replace_si i m nsi ctx =
    match ctx with
      [] -> typing_error dummyinfo "Replace si"
    | (v, (osi, ty)) :: vs -> if i = m then
	(v, (Some nsi, ty)) :: vs
      else
	(v, (osi, ty)) :: replace_si (i+1) m nsi vs

  let update_ctx ctx v si =
    {ctx with var_ctx = replace_si 1 v.v_index si ctx.var_ctx}

  (* Cannot fail *)
  let get_var_bi ctx v =
    return @@ snd (access_var ctx v.v_index)

  let get_var_si ctx v =
    get_var_bi ctx v                        >>= fun (si, _) ->
    return @@ si

  (* XXX: Uhhhhg, shouldn't need to construct this *)
  let get_binder_si ctx bi =
    let v = { dvi with
      v_index = 1;
      v_name = bi.b_name;
      v_type = bi.b_type;
      v_size = bi.b_size;
    } in
    get_var_bi ctx v                        >>= fun (si, _) ->
    return @@ si

  let get_var_ty ctx v =
    get_var_bi ctx v                            >>= fun (_, ty) ->
    return @@ ty

  (* Main context functions. The merge ones take care of application,
     the eq ones take care of the ampersand rule *)

  (* It should be modified when we feature subtyping *)
  let merge_elem i r v1 v2 =
    match (v1, v2) with
    | (vn1, (None,     ty1)), (vn2, (None,     _ty2)) ->
      assert (vn1 = vn2);
      (* XXX: This is a speed up, must verify *)
      (* check_type_eq i ty1 ty2 >> *)
      return @@ (vn1, (None, ty1))

    | (vn1, (Some si1, ty1)), (vn2, (None,     _ty2)) ->
      assert (vn1 = vn2);
      (* XXX: This is a speed up, must verify *)
      (* check_type_eq i ty1 ty2 >> *)
      return @@ (vn1, (Some si1, ty1))

    | (vn1, (None,     ty1)), (vn2, (Some si2, _ty2)) ->
      assert (vn1 = vn2);
      (* XXX: This is a speed up, must verify *)
      (* check_type_eq i ty1 ty2 >> *)
      return @@ (vn1, (Some (SiMult(r, si2)), ty1))

    | (vn1, (Some si1, ty1)), (vn2, (Some si2, ty2)) ->
      assert (vn1 = vn2);
      check_type_eq i ty1 ty2 >>
      return @@ (vn1, (Some (SiAdd (si1, SiMult(r, si2))), ty1))

  (* Note that in this implementation, both contexts are exactly of the
     same length *)
  (* This code can be improved and refactored a lot *)
  let rec merge_varctx i r ctx1 ctx2 =
    match (ctx1, ctx2) with
      ([], [])               -> return []
    | (v1 :: c1), (v2 :: c2) ->
      merge_elem   i r v1 v2         >>= fun v ->
      merge_varctx i r c1 c2         >>= fun l ->
      return (v :: l)
    | ([], _)
    | (_, []) -> assert false

  (* res = ctx1 + r * ctx2 *)
  let merge_ctx i ctx1 r ctx2 =
    (* Apply substs to ctx *)
    normalize_ctx ctx1                           >>= fun ctx1 ->
    normalize_ctx ctx2                           >>= fun ctx2 ->

    ty_debug3 i "--- Merging ctx1: @[%a@]" Print.pp_context ctx1;
    ty_debug3 i "--- Merging ctx2: @[%a@]" Print.pp_context ctx2;

    merge_varctx i r ctx1.var_ctx ctx2.var_ctx   >>= fun ctx ->

    let r_ctx = {ctx1 with var_ctx = ctx} in
    ty_debug3 i "--- Merge result: @[%a@]" Print.pp_context r_ctx;

    return r_ctx

  (* eq_* are unused for now *)
  let eq_elem i v1 v2 =
    match (v1, v2) with
    | (vn1, (None,     ty1)), (vn2, (None,     ty2)) ->
      assert (vn1 = vn2);
      check_type_eq i ty1 ty2 >>
      return @@ (vn1, (None, ty1))

    | (vn1, (Some si1, ty1)), (vn2, (None,     ty2)) ->
      assert (vn1 = vn2);
      check_type_eq i ty1 ty2 >>
      return @@ (vn1, (Some si1, ty1))

    | (vn1, (None,     ty1)), (vn2, (Some si2, ty2)) ->
      assert (vn1 = vn2);
      check_type_eq i ty1 ty2 >>
      return @@ (vn1, (Some si2, ty1))

    | (vn1, (Some si1, ty1)), (vn2, (Some si2, ty2)) ->
      assert (vn1 = vn2);
      check_type_eq i ty1 ty2   >>
      post_si_eq    i si1 si2   >>
      return @@ (vn1, (Some si1, ty1))

  let rec eq_varctx i ctx1 ctx2 =
    match (ctx1, ctx2) with
      ([], [])               -> return []
    | (v1 :: c1), (v2 :: c2) ->
      eq_elem   i v1 v2         >>= fun v ->
      eq_varctx i c1 c2         >>= fun l ->
      return (v :: l)
    | ([], _)
    | (_, []) -> assert false

  let eq_ctx i ctx1 ctx2 =
    (* Apply substs to ctx *)
    normalize_ctx ctx1                           >>= fun ctx1 ->
    normalize_ctx ctx2                           >>= fun ctx2 ->

    eq_varctx i ctx1.var_ctx ctx2.var_ctx        >>= fun ctx ->
    return ({ctx1 with var_ctx = ctx})

  (* When we try to compare two contexts, we must weaken them *)
  let weaken_elem i v1 v2 =
    match (v1, v2) with
    | (vn1, (None,     ty1)), (vn2, (None,     ty2)) ->
      assert (vn1 = vn2);
      check_type_eq i ty1 ty2                        >>
      return @@ (vn1, (None, ty1))

    | (vn1, (Some si1, ty1)), (vn2, (None,     ty2)) ->
      assert (vn1 = vn2);
      check_type_eq i ty1 ty2                        >>
      return @@ (vn1, (Some si1, ty1))

    | (vn1, (None,     ty1)), (vn2, (Some si2, ty2)) ->
      assert (vn1 = vn2);
      check_type_eq i ty1 ty2                        >>
      return @@ (vn1, (Some si2, ty1))

    | (vn1, (Some si1, ty1)), (vn2, (Some si2, ty2)) ->
      let f_si = si_fresh ()                         in
      assert (vn1 = vn2);
      check_type_eq i ty1  ty2                       >>
      post_si_ge    i f_si si1                       >>
      post_si_ge    i f_si si2                       >>
      return @@ (vn1, (Some f_si, ty1))

  let rec weaken_varctx i ctx1 ctx2 =
    match (ctx1, ctx2) with
      ([], [])               -> return []
    | (v1 :: c1), (v2 :: c2) ->
      weaken_elem   i v1 v2         >>= fun v ->
      weaken_varctx i c1 c2         >>= fun l ->
      return (v :: l)
    | ([], _)
    | (_, []) -> assert false

  let weaken_ctx i ctx1 ctx2 =
    (* Apply substs to ctx *)
    normalize_ctx ctx1                           >>= fun ctx1 ->
    normalize_ctx ctx2                           >>= fun ctx2 ->

    weaken_varctx i ctx1.var_ctx ctx2.var_ctx    >>= fun ctx ->
    return ({ctx1 with var_ctx = ctx})

end

open TypeMonad
open TypeConstraint
open TypeCtx

(**********************************************************************)
(* Main routine                                                       *)
(**********************************************************************)
let rec type_of t ctx =

  ty_debug (tmInfo t) "--> Enter type_of: @[%10a@]"
    (Print.limit_boxes Print.pp_term) t;

  (match t with
  (* Variables *)
    TmVar(i, v)  ->
      get_var_si ctx v                       >>= fun si_v ->

      (* Possibly generate a si var for the variable and update the ctx *)
      let si_n  = si_fresh_nb si_v           in
      post_si_ge i si_n (SiConst 1.)         >>
      let ctx_v = update_ctx ctx v si_n      in

      (* Now the HM part *)
      get_var_ty ctx_v v                     >>= fun ty_v     ->

      gen_type_inst ty_v                     >>= fun ty_vi ->
      return @@ (ty_vi, ctx_v)

  (* The primitive noise operator is a float-indexed family of noise
     funtions, that's why it needs special treatment here!

     Type checking never fails, at least here *)
  | TmFuzz(_, n) ->
   (* fuzz (e : num) (n :[e] num) : fuzzy num *)
    return @@ (TyLollipop(TyPrim PrimNum, SiConst n,
			  TyPrim1(Prim1Fuzzy,(TyPrim PrimNum))), ctx)

  (* Primitive terms *)
  | TmPrim(_, pt) ->
    return @@ (type_of_prim pt, ctx)

  (* Abstraction and Applicacion *)

  (* λ (x :[si] tya_x) : tya_tm { tm } *)
  (* XXX: Review the stuff about optional type annotations *)
  | TmAbs(i, b_x, osity_x, otya_tm, tm) ->

    (* Important: we introduce a "weaking" sensitivity variable only
       when there is no annotation *)
    let (sia_x, tya_x) =
      match osity_x with
      | None         -> (si_fresh(), ty_fresh ())
      | Some(si, ty) -> (si, ty)
        (* The constraint solver should take care of this *)
	(* if si = SiConst(0.0) then *)
	(*   typing_error i "Cannot use an annotation of 0 sens" *)
	(* else *)
    in

    with_extend_ctx ctx b_x.b_name tya_x (         fun ctx_tm ->
      type_of tm ctx_tm                            >>= fun (ty_tm, ctx_tm') ->

      check_type_inst_op i otya_tm ty_tm           >>
      normalize (ty_tm, ctx_tm')
    )                                              >>= fun ((si_x, ty_x), ty_tm, ctx') ->

    (* The annotation should be an instance of the returned type *)
    (* This always happens but for annotations using poly variables *)
    check_type_inst i tya_x ty_x                   >>

    (* The annotated sens should be greater than the actual one *)
    post_si_ge_m i sia_x si_x                      >>

    begin
      (match si_x with
      | None    -> ()
      | Some si ->
	ty_info2 i "### Inferred sensitivity %a for parameter %a"
          Print.pp_si si Print.pp_binfo b_x);

    normalize (TyLollipop(ty_x, sia_x, ty_tm), ctx')
    end

  (* tm1 !ᵢβ → α, tm2: β *)
  | TmApp(i, tm1, tm2)                             ->

    type_of tm1 ctx                                 >>= fun (ty1, ctx_1)  ->
    type_of tm2 ctx                                 >>= fun (ty2, ctx_2)  ->

    (* Checks that ty1 has shape !ᵢβ → α, and that ty2 is and instance of β.
       Returns α and the sensitivity inside ty1 *)
    check_app i ty1 ty2                             >>= fun (tya, r) ->

    (* For now we use the sensitivity that is in the type, which will
       be usually the one provided by the user. We do want to infer
       sensitivities for applications yet *)
    merge_ctx i ctx_1 r ctx_2                       >>= fun n_ctx ->

    (* Substitute variables and return the type of the appliaction *)
    return @@ (tya, n_ctx)

  (************************************************************)
  (* Identical to app + lam, but this rule exists in order to handle
     HM polymorphism!

     Do we need the value restricition for soundness? So far we have
     no references so we should be safe AFAICS.
  *)
  (* let x [: otya_x] = tm in e *)
  | TmLet(i, x, otya_x, tm_x, e)                         ->

    (* x may be annotated with a type scheme of the form ∀ x . T[x],
       where x is encoded using DeBruijn. Thus, we need to "open" all
       the annotations in tm_x before proceeding with type checking.
    *)

    let tm_x' = open_annot otya_x tm_x                in
    type_of tm_x' ctx                                 >>= fun (ty_x, ctx_x)  ->

    (* Not entirely right yet. Should ∀ x y, x ≺ ∀ x, x hold?

       Review p. 14 (Fig 4) of Peyton-Jones, Vytiniotis, Weirich, and
       Shields for the likely right fix to this situation on how to
       improve our type_inst
    *)

    let gty_x = generalize_type ctx_x ty_x in
    check_type_inst_op i otya_x gty_x                 >>

    let e_ty = Option.default gty_x otya_x            in

    ty_info2 i "### Inferred type for binder %a is %a"
      Print.pp_binfo x Print.pp_type e_ty;

    with_extend_ctx ctx x.b_name e_ty (                   fun ctx_x         ->
      type_of e ctx_x
    )                                                 >>= fun ((si_x, _ty_x), ty_e, ctx_e)  ->

    (* Optimization: If x was not used, we may as well don't lose time
       merging a context multiplied by 0 *)
    let si_x' =
      begin
	match si_x with
	| None       -> si_fresh ()
	| Some si_x' -> si_x'
      end
    in
    merge_ctx i ctx_e si_x' ctx_x            >>= fun ctx_f ->

    ty_info2 i "### Assigning sensitivity %a to binding %a"
      Print.pp_si si_x' Print.pp_binfo x;
    return @@ (ty_e, ctx_f)

  (* Tensor and & *)
  | TmAmpersand(i, tm1, tm2)      ->

    type_of tm1 ctx                                   >>= fun (ty1, ctx_1) ->
    type_of tm2 ctx                                   >>= fun (ty2, ctx_2) ->

    weaken_ctx i ctx_1 ctx_2                          >>= fun n_ctx ->

    return @@ (TyAmpersand(ty1, ty2), n_ctx)

  (************************************************************)
  (* Data type manipulation *)
  | TmFold(i, ty, tm)               ->

    check_mu_shape i ty             >>= fun ty_unf ->

    type_of tm ctx                  >>= fun (ty_m, ctx_m) ->

    check_type_unif i ty_unf ty_m   >>
    normalize (ty, ctx_m)

  | TmUnfold (i, tm)                ->

    type_of tm ctx                  >>= fun (ty_m, ctx_m) ->

    check_mu_shape i ty_m           >>= fun ty_unf ->
    normalize (ty_unf, ctx_m)

  (* Aliases *)
  | TmTypedef(i, n, tdef_ty, tm)    ->

    ty_debug3 i "Calling typedef %a = %a on term %a" Print.pp_binfo n Print.pp_type tdef_ty Print.pp_term tm;

    (* We just substitute the type for the variable. *)
    let tm'         = term_ty_subst 1 (fun _ -> tdef_ty) tm   in
    type_of tm' ctx

  (************************************************************)
  (* Not core rules         *)
  (* Candidates for removal *)

  (* let (x,y) = e in e' *)
  | TmTensDest(i, x, y, e, t) ->

    type_of e ctx                                     >>= fun (ty_e, ctx_e)  ->
    check_tensor_shape i ty_e                         >>= fun (ty_x, ty_y)   ->

    with_extend_ctx_2 ctx x.b_name ty_x y.b_name ty_y (   fun l_ctx          ->
      type_of t l_ctx
    )                                                 >>= fun (ty_t, (si_x, si_y), ctx_t) ->

    (* Optimization: If the argument of the let was not used, we may
       as well don't lose time merging ctx_e which is to be multiplied
       by 0
    *)
    let return = normalize in
    let (si_x', si_y') =
      begin match si_x, si_y with
	| None, None             -> ty_warning i "### No use of the parameters in let, Zero context!";
                                    (si_fresh (), si_fresh ())
	| Some si_x', None       -> (si_x', si_fresh ())
	| None, Some si_y'       -> (si_fresh (), si_y')
	| Some si_x', Some si_y' -> (si_x', si_y')
      end
    in
    let f_si = si_fresh ()                in
    post_si_ge i f_si si_x'               >>
    post_si_ge i f_si si_y'               >>
    merge_ctx i ctx_t f_si ctx_e          >>= fun n_ctx ->
    return @@ (ty_t, n_ctx)

  (* case e of inl(x) => e_l | inr(y) e_r *)
  | TmUnionCase(i, e, b_x, e_l, b_y, e_r) ->

    type_of e ctx                                     >>= fun (ty_e, ctx_e)  ->
    check_union_shape i ty_e                          >>= fun (ty_x, ty_y)   ->

    with_extend_ctx ctx b_x.b_name ty_x (                 fun ctx_l ->
      type_of e_l ctx_l
    )                                                 >>= fun ((si_x, _), ty_l, ctx_l) ->

    with_extend_ctx ctx b_y.b_name ty_y (                 fun ctx_r ->
      type_of e_r ctx_r
    )                                                 >>= fun ((si_y, _), ty_r, ctx_r) ->

    weaken_ctx      i ctx_l ctx_r                     >>= fun ctx_f ->

    check_type_unif i ty_l ty_r                       >>

    (* Similar trick than above, optimize out *)
    let return = normalize in
    (* We cannot optimize now due the >0 flaw *)
    let (si_x', si_y') =
      begin
	match si_x, si_y with
	| None, None             -> ty_warning i "### No use of the parameters in case, Zero context!";
                                    (si_fresh (), si_fresh ())
	| Some si_x', None       -> (si_x', si_fresh ())
	| None, Some si_y'       -> (si_fresh (), si_y')
	| Some si_x', Some si_y' -> (si_x', si_y')
      end
    in
    let f_si = si_fresh ()                in
    post_si_ge i f_si si_x'               >>
    post_si_ge i f_si si_y'               >>
    merge_ctx i ctx_f f_si ctx_e          >>= fun n_ctx ->
    ty_info2 i "### Inferred sensitivity %a for case left parameter  %a"
      Print.pp_si si_x' Print.pp_binfo b_x;
    ty_info2 i "### Inferred sensitivity %a for case right parameter %a"
      Print.pp_si si_y' Print.pp_binfo b_y;
    return @@ (ty_l, n_ctx)

  (********************************************************************)
  (* Dependent stuff, TODO *)

  (* Instantiation of TyEVars and contexts:

     The general rule is that we cannot instantiate an existential
     type variable with a type that contains a universal one (bound by
     a forall) if the existential variable was created when that
     variable was not introduced.

     Note that a existential may escape, this means we have to use
     "versioning" of the deBruijn indexes in order to keep track of
     the correct instantiation permissions.

     TODO: Research some literature on this...

     What those rules means for ctx and fresh variable manipulation?

     - Every TyEVar must record the context it was created. How to
       record it is detailed below.

     - The use of a context lenght (given that we are using DeBruijn
     indexes) is not enough to effectively differentiate different
     contexts from the instantiation point of view.

     - A simple structure that works is to "version" the contexts. For
     that purpose, every time we extend a context we'll increment a
     global counter for that position.

     - Thus, the context belonging to an existential type variable is
     a list of integers.

     - An existential type variable can be instantiated with a type
     containing a particular type variable iff the variable unique
     number is a member of its list.

  *)

  (*

    Constraint stores: We need support for real constraint contexts of
    the form Γ ⊢ φ, the constraint store is then a list of such
    judgments [or alternatively a formula (Γ₁ → φ₁) ∧ ⋯ ∧ (Γₙ → φₙ)]

    We also extend the syntax for constraints with universal
    quantification, thus constraints like:

    (s = 0) → ∀ x . x ≥ (s + 1)

    are possible.

    What is the optimal representation for our constraint, which are
    of the form:

    (α₁ → β₁) ∧ ((α₁ ∧ α₂) → β₂) ∧ ((α₁ ∧ α₂ ∧ α₃) → β₃)

      α₁            → β₁  ∧
     (α₁ ∧ α₂)      → β₂) ∧
     (α₁ ∧ α₂ ∧ α₃) → β₃)

    Is rewritten as:

            α₁  →
      β₁ ∧ (α₂  →
      β₂ ∧ (α₃  →
      β₃         ))

    α₁ →
         β₁ ∧ (α₂ →
                    β₂ ∧ (α₃ →
                               β₃))

    Thanks to the lemma:

    [Begin COQ]

    Variable a1 a2 b1 b2 : Prop.
    Lemma l1 : (a1 -> b1) /\ ((a1 /\ a2) -> b2) <->
               (a1 -> (b1 /\ (a2 -> b2))).
    tauto.
    Qed.

    [End COQ]

  *)

  (* Dependent pattern matching *)
  | TmListCase(i, _tm, _ltm, _elem, _list, _si, _rtm) ->
    typing_error i "DFuzz stub: Listcase"

  | TmNatCase(i, _tm, _ltm, _nat, _si, _rtm)          ->
    (*

      We need to extend both the typing context and the constraint
      store context. We then type the branches with the next context.

      thus we need an

      extend_ctx_case : constraint -> binding * kind -> ctx -> ctx

      About the generic rule:

      Think of the generic rule. We may want to use it, as the
      complexity of the rule will be too big otherwise, given that we
      have an implicit si abstraction.

      - basically all the constraints in each branch may use the newly
      learned fact in order to be satisfied.

    *)
    typing_error i "DFuzz stub: NatCase"

  (* Pack/Unpack *)
  | TmPack(i, _, _, _)               ->
    typing_error i "DFuzz stub: Pack"

  | TmUnpack(i, _n, _si, _t1, _t2)   ->
    typing_error i "DFuzz stub: Unpack"

  (* Type/Sensitity Abstraction and Applicacion *)
  | TmSiAbs(i, _bi, _ki, _tm) ->
    (*
      - Introduce the binding into the context.
      - Get the type of tm
      - The type is forall si

      - Note that none of the free variables should be instantiated to
      the same binding. Is that possible?

      We must close and shift the constraint store!

      Pseudocode:

      with_extend_ty_ctx bi ki (type_of tm) >>= fun ty ->

      close_cs !! All the constraints get a forall.

      return TyForall bi ki ty

      Question: Should we check that the kind is correct?
    *)
    typing_error i "DFuzz stub: Sens abstraction"

  | TmSiApp(i, _tm, _si)      ->
    (*
      - Get the type of tm.
      - It should be forall
      - Substitute, this is the new type.

      - What happens with the free variables in si?

        I guess they cannot be instantiated to the fresh binder.

      Pseudocode:

      type_of tm              >>= fun ty   ->
      check_forall_shape ty   >>= fun ty_i ->
      return (ty_subst ty_i si)

    *)
    typing_error i "DFuzz stub: Sens abstraction"
) >>= fun (ty, ctx) ->

  (* We limit pp_term *)
  ty_debug (tmInfo t) "<-- Exit type_of: @[%a@] with type @[%a@]"
    (Print.limit_boxes Print.pp_term) t Print.pp_type ty;

  ty_debug2 (tmInfo t) "<-- Context: @[%a@]" Print.pp_context ctx;

  return (ty, ctx)

open Format
open Print

let pp_tyerr ppf s = match s with
  | SensError cs          -> fprintf ppf "Cannot satisfy constraint %a"   Constraint.pp_si_cstr_elem cs
  | MoreGeneral(ty1, ty2) -> fprintf ppf "%a is not more general than %a" pp_type ty1 pp_type ty2
  | TypeInst(ty1, ty2)    -> fprintf ppf "Type %a is not instance of %a"  pp_type ty1 pp_type ty2
  | TypeMismatch(ty1, ty2)-> fprintf ppf "Cannot unify %a with %a"        pp_type ty1 pp_type ty2
  | CannotApply(ty1, ty2) -> fprintf ppf "Cannot apply %a to %a"    pp_type ty1 pp_type ty2
  | OccursCheck(v, ty)    -> fprintf ppf "Cannot build infinite type %a = %a" pp_vinfo v pp_type ty
  | WrongShape(ty, sh)    -> fprintf ppf "Type %a has wrong shape, expected %s type."  pp_type ty sh

(* Equivalent to run *)
let get_type program =
  let (State f) = type_of program Ctx.empty_context in
  match f (empty_cs, ty_esub_id) with
    Left e        -> typing_error_pp e.i pp_tyerr e.v
  | Right ((cs, _s), (ty, ctx)) -> (cs, (ty, ctx))
