open Ctx
open Syntax
open Constraint

open Support.Pervasive
open Support.Error
open Print

(* Is "easy" to modify this monad in order to do type inference (if
   decidable). Some ideas:

   - infer-shape: Forces the return to be of a particular shape.
   - check_type: Checks against annotations.
   - unify_types: What the name says, akin to the merge operation.
*)

(* Type checking monad for dfuzz: (ctx, cs) -> (ctx, cs, ty) option *)

(* Errors *)
type ty_error_elem =
| SensError
| TypeMismatch

type 'a ty_error =
  Right of 'a
| Left  of ty_error_elem

module TyperMonad = struct

  (* If needed, we could keep the state and push the error inside *)
  type 'a tyState =
    State of (Ctx.context ->
              (Constraint.c_store * 'a) ty_error)

  (* Native @@ already is in ocaml 4.0 *)
  let (@@) x y = x y

  let fail e    = State (fun _ -> Left e)
  let return x  = State (fun _ -> Right ([], x))

  let (>>=) m f = State (fun ctx ->
    let (State mf) = m in
    match mf ctx with
      Left e               -> Left e
    | Right (cs, a) ->
      let State ff = f a in
      begin
	match ff ctx with
	  Left e         -> Left  e
	| Right (cs', y) -> Right ((cs@cs'),y)
      end)

  let (>>) m f = m >>= fun _ -> f

  let get_ctx = State (fun ctx ->
    Right ([], ctx))

  let put_cs cs = State (fun _ ->
    Right (cs, ()))

end

open TyperMonad

(* Monad operations *)

let with_ctx ctx m = State (fun _ ->
  let State f = m in
  f ctx)

let with_extend_ctx v si ty m =
  get_ctx                                 >>= fun ctx ->
  with_ctx (extend_var v (si, ty) ctx) m

let get_var_bi v =
  get_ctx                                 >>= fun ctx ->
  return @@ snd (access_var ctx v.v_index)

let get_var_si v =
  get_var_bi v                            >>= fun (si, _) ->
  return @@ si

let get_var_ty v =
  get_var_bi v                            >>= fun (_, ty) ->
  return @@ ty

(* Scheduled for removal *)
let get_binder_si bi =
  let v = {v_index = 1; v_name = bi.b_name; v_type = bi.b_type; v_size = bi.b_size} in
  get_var_bi v                            >>= fun (si, _) ->
  return @@ si

(* Can fail *)
let check_type_eq _i _ty1 _ty2 =
  return ()

(* XXX we have to constraint the inferred sensitivity to the actual one *)
let check_app_shape _i ty1 _si = return (ty1, ty1)
let check_tensor_shape _i ty1 = return (ty1, ty1)

(* Check equatily against an optional annotation *)
let check_type_annot_op i ty oty =
  match oty with
    None     -> return ()
  | Some ty' -> check_type_eq i ty ty' (* This should be inst *)

(* Can fail *)
let post_si_ge si1 si2 =
  match Constraint.post_si_ge [] si1 si2 with
    None     -> fail SensError
  | Some cs' -> put_cs cs'

(* Key function *)
(* x :[si] t -> (x :[s1] t), x :[s2] t) si >= s1 + r s2 *)
let split_elem (v, (si, ty)) r =
  let (s1, s2) = (si_fresh (), si_fresh ()) in
  let term     = SiAdd(s1, SiMult(r, s2))   in
  let v1 = (v, (s1, ty))      in
  let v2 = (v, (s2, ty))      in
  post_si_ge si term          >>
  return (v1, v2)

(* Note that in this implementation, both contexts are exactly of the
   same length *)
let rec split_varctx ctx r =
  match ctx with
    []       -> return ([],[])
  | (v :: l) ->
    split_elem v r                     >>= fun (v1, v2) ->
    split_varctx l r                   >>= fun (l1, l2) ->
    return (v1 :: l1, v2 :: l2)

let split_ctx r ctx =
  split_varctx ctx.var_ctx r                       >>= fun (ctx1, ctx2) ->
  return ({ctx with var_ctx = ctx1},
	  {ctx with var_ctx = ctx2})

(* Takes two computations, with the second depending on the first and
   returns a new one with the original context, but a new, merged
   cs *)
(* let merge_cs (State m1) f = State (fun (ctx, cs) -> *)
(*   let comp1 = m1 (ctx, cs) in *)
(*   match comp1 with *)
(*     Left e1            -> Left e1 *)
(*   | Right (_, cs1, r1) -> *)
(*     begin *)
(*       let (State m2) = f r1          in *)
(*       let comp2      = m2 (ctx, cs1) in *)
(*       match comp2 with *)
(* 	Left e2            -> Left e2 *)
(*       | Right (_, cs2, r2) -> *)
(* 	Right (ctx, cs2, (r1, r2)) *)
(*     end) *)

(* Equivalent to the above ? *)
let merge_cs m1 f =
  m1                        >>= fun r1 ->
  f r1                      >>= fun r2 ->
  return @@ (r1, r2)

let split r m1 m2 =
  get_ctx                           >>= fun ctx ->
  split_ctx r ctx                   >>= fun (ctx1, ctx2) ->
  merge_cs (with_ctx ctx1 m1)
           (fun x -> with_ctx ctx2 (m2 x))

let equate m1 m2 =
  m1 >>= fun r1 ->
  m2 >>= fun r2 ->
  return (r1, r2)

(* Polymorhpic instantiation *)

(* let check_type_inst = fun (ctx, cs) -> *)

(* the sensitivies here are handled by the caller *)


(* Horrible debug code *)
(* No String.map until Ocaml 4.0 *)
let rec s_map f s =
  for i = 0 to (String.length s) -1 do
    String.set s i (f (String.get s i))
  done; s

(**********************************************************************)
(* Debug code *)
let rec type_of t =
  (* Horrible debug code *)
  let term_full = s_map (fun x -> if x = '\n' then ' ' else x) (string_of_term t)            in
  let cut_n     = 50                                                                         in
  let cut_s     = if String.length term_full > cut_n then cut_n else String.length term_full in
  let term = (String.sub term_full 0 cut_s) in
  get_ctx >>= fun ctx ->
  if (match t with | TmPrim(_,_) -> false | _ -> true) then
    begin
      pr "=====> ENTER typeof: " ; pr term ; pr "\n";
      pr "=====> Context:\n" ; dumpctx ctx
    end;
(* End Debug code *)
(**********************************************************************)

(* let rec type_of t = *)
(  match t with
  (* Variables *)
    TmVar(_i, v)  ->
      get_var_si v                        >>= fun si ->
      post_si_ge si (SiConst 1.)          >>

      (* XXX: Poly: Instantiate type *)
      get_var_ty v

  (* Primitive terms *)
  | TmPrim(_, pt) ->
    return @@ type_of_prim pt

  (* Abstraction and Applicacion *)

  (* λ (bi :[si] a_ty) : r_tyo { tm } *)
  | TmAbs(i, bi, si, a_ty, r_tyo, tm) ->

    with_extend_ctx bi.b_name si a_ty (
      type_of tm                                   >>= fun ty_tm ->
      check_type_annot_op i ty_tm r_tyo            >>
      return ty_tm
    )                                              >>= fun ty_tm ->

    return @@ TyLollipop(a_ty, si, ty_tm)

  (* tm1 !ᵢβ → α, tm2: β *)
  | TmApp(i, tm1, tm2)                ->

    let r = si_fresh ()                          in
    split r (type_of tm1)
            (fun _ -> type_of tm2)               >>= fun (ty1, ty2) ->

    check_app_shape i ty1 r                      >>= fun (tyb, tya) ->
    check_type_eq i tyb ty2                      >>

    return @@ tya

  (* let bi [: oty] = tm1 in e *)
  | TmLet(i, bi, oty, tm1, e)            ->
    let r = si_fresh ()                          in

    split r (type_of tm1)
            (fun ty1 ->
	      with_extend_ctx bi.b_name r ty1 (
		type_of e
	      )
	    )                                    >>= fun (ty1, tye) ->

    check_type_annot_op i ty1 oty                >>
    return tye

  (* Both eliminators are candidates for removal *)

  (* let (x,y) = e in e' *)
  | TmTensDest(i, x, y, e, t) ->
    let r = si_fresh ()                         in

    split r (type_of e)
            (fun tye ->
	      check_tensor_shape i tye          >>= fun (tyx, tyy) ->
	      with_extend_ctx x.b_name r tyx (
		with_extend_ctx y.b_name r tyy (
		  type_of t
		)
	      )
	    )                                   >>= fun (ty1, tye) ->
    return tye

  (* case e of inl(x) => e_l | inr(y) e_r *)
  (* | TmUnionCase(i, e, x, e_l, y, e_r) -> *)
  (*   return @@ (TyPrim PrimUnit) *)

  (* Tensor and & *)
  | TmAmpersand(_i, tm1, tm2)      ->
    equate (type_of tm1)
           (type_of tm2)                   >>= fun (ty1, ty2) ->

    return @@ TyAmpersand(ty1, ty2)

  (* (\* Data type manipulation *\) *)
  (* | TmFold(i, ty, tm)            -> *)
  (*   return @@ (TyPrim PrimUnit) *)

  (* | TmUnfold (i, tm)            -> *)
  (*   return @@ (TyPrim PrimUnit) *)

  (* (\* Aliases *\) *)
  (* | TmTypedef(_, n, ty, tm)    -> *)
  (*   return @@ (TyPrim PrimUnit) *)

  (********************************************************************)
  (* Dependent stuff, for later *)

  (* Dependent pattern matching *)
  | TmListCase(i, _tm, _ltm, _elem, _list, _si, _rtm) ->
    error i "No support (yet) for dependent types"
  | TmNatCase(i, _tm, _ltm, _nat, _si, _rtm)          ->
    error i "No support (yet) for dependent types"

  (* Pack/Unpack *)
  | TmPack(i, _, _, _)               ->
    error i "No support (yet) for dependent types"

  | TmUnpack(i, _n, _si, _t1, _t2)   ->
    error i "No support (yet) for dependent types"

  (* Type/Sensitity Abstraction and Applicacion *)
  | TmSiAbs(i, _bi, _ki, _tm) -> error i "No support (yet) for dependent types"
  | TmSiApp(i, _tm, _si)      -> error i "No support (yet) for dependent types"
) >>= fun ty ->
  pr "****** Exit: "; pr term; pr "\n";
  pr "****** Return type: ";
  pr (string_of_type ty); pr "\n";
  (* get_cs >>= fun cs -> *)
  (* pr "****** Constraint store: \n"; *)
  (* pr (string_of_si_cs cs); *)
  (* pr "\n"; *)
  pr "****** Constraint store for z3: \n";
      pr (string_of_si_cs_z3 cs);
      pr "\n";
  return ty

(* Equivalent to run *)
let get_type program =
  let (State f) = type_of program in
  match f Ctx.empty_context with
    Left _e            -> error dummyinfo "Error in typing"
  | Right (cs, ty) -> (cs, ty)


