(*
  Fuzz compiler.

  Copyright: The University of Pennsylvania 2012
  License: ...

  Z3 output

*)

(* For now we only support pretty printing *)

open Syntax
open Constraint

open Format

(***********************************************************************)
(* Pretty printing for Z3 (SMT Output) *)

let rec pp_si_z3 ppf s =
  match s with
  (* This will ocurr in dependent types *)
  | SiVar(_v)        -> fprintf ppf "Universal quantified variable not supported for Z3"
  | SiIntConst(i)    -> fprintf ppf "%d" i
  (* *)
  | SiEVar(v)        -> fprintf ppf "S%d" v.v_index
  | SiInfty          -> fprintf ppf "infty"
  | SiConst(flt)     -> fprintf ppf "(mk-r %.1f)" flt
  | SiAdd (si1, si2) -> fprintf ppf "(inf-sum %a %a)" pp_si_z3 si1 pp_si_z3 si2
  | SiMult(si1, si2) -> fprintf ppf "(inf-prod %a %a)" pp_si_z3 si1 pp_si_z3 si2

let rec pp_list_z3 pp fmt l = match l with
    []         -> fprintf fmt ""
  | csx :: csl -> fprintf fmt "(assert %a)\n%a" pp csx (pp_list_z3 pp) csl

let pp_si_cstr_elem_z3 fmt si_cstr = match si_cstr with
    CSiEq(_, si1, si2) -> fprintf fmt "(inf-eq %a %a)" pp_si_z3 si1 pp_si_z3 si2
  | CSiGe(_, si1, si2) -> fprintf fmt "(inf-geq %a %a)" pp_si_z3 si1 pp_si_z3 si2

let pp_si_cstr_z3 = (pp_list_z3 pp_si_cstr_elem_z3)

let cs_elem_fv cs = match cs with
  | CSiEq(_, si1, si2) -> VarSet.union (si_efv si1) (si_efv si2)
  | CSiGe(_, si1, si2) -> VarSet.union (si_efv si1) (si_efv si2)

let rec cs_free_vars cs = match cs with
  | []          -> VarSet.empty
  | (cs :: csl) -> VarSet.union (cs_elem_fv cs) (cs_free_vars csl)

(* *)
let pp_var_decl_z3 fmt varset =
  VarSet.iter (fun var ->
    fprintf fmt "(declare-const S%d (Pair Bool Real))\n" var.v_index
  ) varset

let pp_si_store_z3 fmt cs =
  fprintf fmt "@[<v>%a@]" pp_si_cstr_z3 cs

(* TODO:
   - Generate the set of free variables of the constraints.
   - Finish the pretty print.
*)

let z3_prelude =
"(declare-datatypes (T1 T2) ((Pair (mk-pair (first T1) (second T2)))))
(declare-const infty (Pair Bool Real))
(assert (= infty (mk-pair true 0.)))
(define-fun inf-sum ((p1 (Pair Bool Real)) (p2 (Pair Bool Real))) (Pair Bool Real)
  ( ite
     (first p1)
     p1
     (ite
       (first p2)
       p2
       (mk-pair false (+ (second p1) (second p2)))
      )
  )
)
(define-fun inf-prod ((p1 (Pair Bool Real)) (p2 (Pair Bool Real))) (Pair Bool Real)
  ( ite
     (first p1)
     	(ite
		(and (not (first p2)) (= (second p2) 0.))
		(mk-pair false 0.)
		p1
	)
     (ite
        (first p2)
          (ite (= 0. (second p1))
            p1
            p2
          )
          (mk-pair false (* (second p1) (second p2)))
     )
  )
)
(define-fun inf-eq ((p1 (Pair Bool Real)) (p2 (Pair Bool Real))) Bool
  ( ite
     (first p1)
     	(first p2)
     (ite
	(first p2)
          (first p1)
        (= (second p1) (second p2))
     )
   )
)
(define-fun inf-geq ((p1 (Pair Bool Real)) (p2 (Pair Bool Real))) Bool
  ( ite
     (first p1)
     	true
     (ite
	(first p2)
        false
        (>= (second p1) (second p2))
     )
   )
)
(define-fun mk-r ((r Real)) (Pair Bool Real)
(mk-pair false r))"

let z3_appendix =
"(check-sat)
(get-model)\n
"
let gen_z3_constraint fmt cs =
  let cs_fv = cs_free_vars cs       in
  fprintf fmt "%s\n%a\n%a\n%s@." z3_prelude pp_var_decl_z3 cs_fv pp_si_store_z3 cs z3_appendix


