(*
  Fuzz compiler.

  Copyright: The University of Pennsylvania 2012
  License: ...

  Z3 output

*)

(* For now we only support pretty printing *)

open Syntax
open Constraint

open Format

(***********************************************************************)
(* Pretty printing for Z3 (Python output)*)

let rec pp_si_z3_py ppf s =
  match s with
  (* This will ocurr in dependent types *)
  | SiVar(_v)        -> fprintf ppf "Universal quantified variable not supported for Z3"
  | SiIntConst(i)    -> fprintf ppf "%d" i
  (* *)
  | SiEVar(v)        -> fprintf ppf "s%d" v.v_index
  | SiInfty          -> fprintf ppf "Inf()"
  | SiConst(flt)     -> fprintf ppf "MyRVal(%.1f)" flt
  | SiAdd (si1, si2) -> fprintf ppf "(%a + %a)" pp_si_z3_py si1 pp_si_z3_py si2
  | SiMult(si1, si2) -> fprintf ppf "(%a * %a)" pp_si_z3_py si1 pp_si_z3_py si2

let rec pp_list_z3_py pp fmt l = match l with
    []         -> fprintf fmt ""
  | [cs]       -> fprintf fmt "%a" pp cs
  | csx :: csl -> fprintf fmt "%a,@,%a" pp csx (pp_list_z3_py pp) csl

let pp_si_cstr_elem_z3_py fmt si_cstr = match si_cstr with
    CSiEq(_, si1, si2) -> fprintf fmt "%a == %a" pp_si_z3_py si1 pp_si_z3_py si2
  | CSiGe(_, si1, si2) -> fprintf fmt "%a >= %a" pp_si_z3_py si1 pp_si_z3_py si2

let pp_si_cstr_z3 = (pp_list_z3_py pp_si_cstr_elem_z3_py)

(* *)
let pp_var_decl_z3_py fmt varset =
  VarSet.iter (fun var ->
    fprintf fmt "s%d = MyR('%s')\n" var.v_index var.v_name
  ) varset

(* Loris rightly notes that doing s%d.val >= 0.0 will generate less
   constraints, we keep this approach in order to respect
   encapsulation *)
let pp_var_geq0_z3_py fmt varset =
  VarSet.iter (fun var ->
    fprintf fmt "s%d > MyRVal(0.0),@," var.v_index
  ) varset

let pp_vars_geq0_z3_py fmt cs =
  fprintf fmt "@[<v>%a@]" pp_var_geq0_z3_py cs

let pp_si_store_z3_py fmt (cs_fv, cs) =
  fprintf fmt "@[<v>%a@]@[<v>%a@]"
    pp_vars_geq0_z3_py cs_fv
    pp_si_cstr_z3 cs

let pp_var_j_name fmt v =
  fprintf fmt "s%d" v.v_index

let pp_var_array =
  pp_list_z3_py pp_var_j_name

let pp_si_cs_string fmt cs =
  fprintf fmt "\"%a\"" Constraint.pp_si_cstr_elem_fi cs

let pp_si_store_z3_string =
  pp_list_z3_py pp_si_cs_string

(* TODO: - It crashes if trying to miminize a variable that is not the
   constraint store!
*)

let z3_py_prelude =
"# coding=utf-8

from z3     import *
from z3rinf import *
"

open Support

(* We may have declared variables that don't appear in the constraint
   store, this will cause a crash on Z3 if trying to minimize them *)
let cs_elem_fv cs = match cs with
  | CSiEq(_, si1, si2) -> VarSet.union (si_efv si1) (si_efv si2)
  | CSiGe(_, si1, si2) -> VarSet.union (si_efv si1) (si_efv si2)

let rec cs_free_vars cs = match cs with
  | []          -> VarSet.empty
  | (cs :: csl) -> VarSet.union (cs_elem_fv cs) (cs_free_vars csl)


let gen_z3_py_constraint fmt cs =
  (* XXX: This is not fully tested and should be considered
     experimental, it will break the minimization list for one *)
  (* let cs     = Constraint.simpl_subst cs                   in *)
  let cs_fv  = cs_free_vars cs                             in
  let cs_min = VarSet.filter (fun v -> v.v_minimize) cs_fv in
  fprintf fmt "%s@\n%a@\nsvars = [@[%a@]]@\ninteresting_svars = [@[%a@]]@\ncs_string = [@[<v>%a@]]@\nconstraints = [@[%a@]]@\nminimize = [@[%a@]]@."
    z3_py_prelude
    pp_var_decl_z3_py cs_fv
    pp_var_array (VarSet.elements cs_fv)
    pp_var_array (VarSet.elements (VarSet.filter (fun v -> v.v_interesting) cs_fv))
    pp_si_store_z3_string cs
    pp_si_store_z3_py (cs_fv, cs)
    pp_var_array (VarSet.elements cs_min)
